# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:41:54 2020

@author: xavier.mouy

"""
import sys
sys.path.append(r"C:\Users\xavier.mouy\Documents\GitHub\ecosound")  # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.detection.detector_builder import DetectorFactory
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
from ecosound.classification.classification import Classifier

import ecosound.core.tools
import time
import pickle
import copy

## Input paraneters ##########################################################

single_channel_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Herring_DFO\5042.200306203002.wav"
#single_channel_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\datasets\DFO_snake-island_rca-in_20190410\noise\1342218252.190430230159.wav"
#single_channel_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\problematic_files\1342218252.190415230156.wav"
#model_filename = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\problematic_files\RF50_model_20201208T223420.sav'
#model_filename = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\problematic_files\XGBoost_model_20201209T132812.sav'
model_filename = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\problematic_files\RF50_model_20201209T134646.sav'

# Spectrogram parameters
frame = 0.02 #0.0625
nfft = 0.02 # 4096
step = 0.005 # 5
fmin = 500
fmax = 5000
window_type = 'hann'


# start and stop time of wavfile to analyze
t1 = 0#141#24
t2 = 300#167#40
## ###########################################################################
tic = time.perf_counter()

# load audio data
sound = Sound(single_channel_file)
sound.read(channel=0, chunk=[t1, t2], unit='sec', detrend=True)

# Calculates  spectrogram
spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='sec')
spectro.compute(sound, dB=True, use_dask=True, dask_chunks=100) #40

# Crop unused frequencies
spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)
spectro1 = copy.deepcopy(spectro)

# Denoise
#spectro.denoise('median_equalizer', window_duration=3, use_dask=True, dask_chunks=(2048,1000), inplace=True)
spectro.denoise('median_equalizer', window_duration=3, use_dask=True, dask_chunks=(50,50000), inplace=True)
spectro2 = copy.deepcopy(spectro)

# Detector
detector = DetectorFactory('BlobDetector', kernel_duration=0.1, kernel_bandwidth=500, threshold=10, duration_min=0.05, bandwidth_min=40)
#detections = detector.run(spectro, use_dask=True, dask_chunks=(2048,1000), debug=False)
detections = detector.run(spectro, use_dask=True, dask_chunks=(4096,50000), debug=False)

# Measurements
spectro_features = MeasurerFactory('SpectrogramFeatures', resolution_time=0.001, resolution_freq=0.1, interp='linear')
measurements = spectro_features.compute(spectro,
                                        detections,
                                        debug=False,
                                        verbose=False,
                                        use_dask=True)


# detections.insert_values(audio_file_name='1342218252.190430230159')
# detections.insert_values(audio_file_dir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\datasets\DFO_snake-island_rca-in_20190410\noise')
# detections.insert_values(audio_file_extension='.wav')
# detections.insert_values(label_class='NN')
# detections.to_pamlab(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\datasets\DFO_snake-island_rca-in_20190410\noise\annotations", outfile='1342218252.190430230159.wav annotations.log', single_file=True)


# # Classification
# loaded_model = pickle.load(open(model_filename, 'rb'))
# features = loaded_model['features']
# model = loaded_model['model']
# Norm_mean = loaded_model['normalization_mean']
# Norm_std = loaded_model['normalization_std']
# classes_encoder = loaded_model['classes']

# data = measurements.data
# X = data[features]
# X = data[features]
# X = (X-Norm_mean)/Norm_std
# pred_class =list(model.predict(X))
# pred_prob = model.predict_proba(X)
# #pred_prob = pred_prob[:,1]
# pred_prob = pred_prob[range(0,len(pred_class)),pred_class]
# #relabel
# for index, row in classes_encoder.iterrows():
#     pred_class = [row['label'] if i==row['ID'] else i for i in pred_class]
# # update measuremnets
# data['label_class'] = pred_class
# data['confidence'] = pred_prob
# data_fish = data[data['label_class']=='FS']
# data_noise = data[data['label_class']=='NN']
# classif_fish = copy.deepcopy(measurements)
# classif_fish.data = data_fish
# classif_noise = copy.deepcopy(measurements)
# classif_noise.data = data_noise



# # Plot
graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=5000)
# graph.add_data(sound)
# #graph.add_annotation(classif_fish, panel=0,color='red', label='Fish', tag=True)
# graph.add_data(spectro1)
# graph.add_data(spectro2)
graph.add_data(spectro2)
# graph.add_data(spectro2)
graph.add_annotation(detections, panel=0,color='red', label='Detections')
# graph.add_annotation(classif_fish, panel=4,color='red', label='Fish', tag=True)
# graph.add_annotation(classif_noise, panel=4,color='blue', label='Noise',tag=True)
# graph.colormap = 'binary'
graph.colormap = 'jet'
graph.show()



# classifier = Classifier()
# classifier.load_model(model_filename)
# classif2 = classifier.classify(measurements)

# # Plot
# graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000)
# graph.add_data(sound)
# #graph.add_annotation(classif_fish, panel=0,color='red', label='Fish', tag=True)
# graph.add_data(spectro1)
# graph.add_data(spectro2)
# graph.add_data(spectro2)
# graph.add_data(spectro2)
# graph.add_annotation(detections, panel=3,color='black', label='Detections')
# graph.add_annotation(classif2, panel=4,color='red', tag=True)
# #graph.add_annotation(classif2, panel=4,color='blue', label='Noise',tag=True)
# graph.colormap = 'binary'
# #graph.colormap = 'jet'
# graph.show()


# toc = time.perf_counter()
# print(f"Executed in {toc - tic:0.4f} seconds")
# # #########################################



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 10:03:39 2020

@author: xavier.mouy
"""
import xarray
import os
import numpy as np

#indir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_In_Oct17_Dec3_2018_67674121'
indir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_Out_April_July2019_1409343536' 

def xr_sort(ds):
    #return ds.sortby(ds.time_min_date,ascending=True)
    ds = ds.sortby(ds.date,ascending=True)
    
    return ds.sortby(ds.date,ascending=True)

def to_drop_vars(ds):
    return [x for x in list(ds.variables) if x not in ['date','audio_file_start_date','time_min_date','label_class','label_subclass','confidence']] 

def fix_duplicates(ds):
    index=range(0,len(ds['date']))
    _, unique_index = np.unique(ds['date'], return_index=True)
    index = np.delete(index,unique_index)
    if len(index)>0:
        rnd = np.round(np.random.rand(len(index))*100)-50
        rnd[rnd==0]=1
        rnd_offset =[np.timedelta64(int(x),'ns') for x in rnd]
        time_ax = ds['date'].data
        time_ax[np.array(index)] = time_ax[index] + rnd_offset
        ds['date'] = time_ax
    #ds.time_min_date = time_ax
    return ds

def shift_max_dates(ds,files_dur_sec):
    time_ax = ds['date'].data    
    max_date = ds['audio_file_start_date'][0].data + np.timedelta64(files_dur_sec,'s')
    index = time_ax >= max_date
    time_ax[np.array(index)] = max_date - np.timedelta64(1,'ns')    
    ds['date'] = time_ax    
    return ds

def preprocess_func(ds):
    files_dur_sec = 1799
    return shift_max_dates(xr_sort(fix_duplicates(ds.drop(to_drop_vars(ds)))),files_dur_sec) 

idx=0
D=np.empty(1)
for dirpath, dnames, fnames in os.walk(indir):
    for f in fnames:
        if f.endswith(".nc"):
            idx+=1
            print(f)
            ds = xarray.open_dataset(os.path.join(indir,f))
            ds = preprocess_func(ds)    
            Tstart = ds['date'].min().data
            Tend = ds['date'].max().data
            print(Tstart, Tend)            
            if idx==1:
                Tpast = Tstart - np.timedelta64(5,'ns')
            if Tpast >= Tstart:
                raise ValueError("ValueError")
            Tpast = Tend
            
            


# -*- coding: utf-8 -*-
"""
Created on Thu Apr 22 10:02:51 2020

@author: xavier.mouy
"""


class BaseClass(object):
    """
    Base class for all measurers.

    All measurers need to inheritate from this BaseClass in order to be built
    by the MeasurerFactory.
    """

    def __init__(self, measurer_name):
        self.measurer_name = measurer_name

    @classmethod
    def is_measurer_for(cls, measurer_name):
        """
        Check measurer name.

        Compare the requested measurer_name with each measurer class
        available.

        Parameters
        ----------
        measurer_name : str
        Name of the detector class.

        Returns
        -------
        bool
        """
        return measurer_name == cls.__name__


def MeasurerFactory(measurer_name, *args, **kwargs):
    """
    Measurer Factory.

    Loads the measurer class defined by measurer_name. Each measurer class must
    be added to the __init__.py file.

    For example :
    from .measurer1 import Measurer1

    where Measurer1 is the name of the measurer class to load from the .py
    file measurer1

    Parameters
    ----------
    measurer_name : str
        Name of the measurer class.
    *args : any
        Input arguments for the measurer selected.
    **kwargs : any
        Keyword arguments for the measurer selected.

    Raises
    ------
    ValueError
        If measurer_name doesn't correspond to any of the measurers available.

    Returns
    -------
    Measurer object

    """
    meas_list = []
    for cls in BaseClass.__subclasses__():
        meas_list.append(cls.__name__)
        if cls.is_measurer_for(measurer_name):
            return cls(measurer_name, *args, **kwargs)
    raise ValueError("Invalid measurer name. Measurers available: "
                     + str(meas_list))



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 20 10:02:51 2020

@author: xavier.mouy
"""


class BaseClass(object):
    """
    Base class for all graphers.

    All graphers need to inheritate from this BaseClass in order to be built
    by the GrapherFactory.
    """

    def __init__(self, grapher_name):
        self.grapher_name = grapher_name

    @classmethod
    def is_grapher_for(cls, grapher_name):
        """
        Check grapher name.

        Compare the requested grapher_name with each grapher class
        available.

        Parameters
        ----------
        grapher_name : str
        Name of the grapher class.

        Returns
        -------
        bool
        """
        return grapher_name == cls.__name__


def GrapherFactory(grapher_name, *args, **kwargs):
    """
    Grapher Factory.

    Loads the grapher class defined by grapher_name. Each grapher class must
    be added to the __init__.py file.

    For example :
    from .grapher1 import Grapher1

    where Grapher1 is the name of the grapher class to load from the .py
    file grapher1

    Parameters
    ----------
    grapher_name : str
        Name of the grapher class.
    *args : any
        Input arguments for the grapher selected.
    **kwargs : any
        Keyword arguments for the grapher selected.

    Raises
    ------
    ValueError
        If grapher_name doesn't correspond to any of the graphers available.

    Returns
    -------
    Grapher object

    """
    detec_list = []
    for cls in BaseClass.__subclasses__():
        detec_list.append(cls.__name__)
        if cls.is_grapher_for(grapher_name):
            return cls(grapher_name, *args, **kwargs)
    raise ValueError("Invalid grapher name. Grapher available: "
                     + str(detec_list))



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 12:24:25 2020

@author: xavier.mouy
"""

import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement
from ecosound.classification.CrossValidation import StratifiedGroupKFold
from ecosound.classification.CrossValidation import RepeatedStratifiedGroupKFold

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import copy
import pickle
import os
from sklearn.model_selection import train_test_split
from sklearn import preprocessing
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import GroupKFold
from sklearn.dummy import DummyClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC
from sklearn.svm import LinearSVC
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import f1_score
from sklearn.preprocessing import LabelEncoder
from sklearn import metrics
from sklearn.decomposition import PCA
from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import RandomUnderSampler
from imblearn.combine import SMOTETomek
from imblearn.under_sampling import TomekLinks
from datetime import datetime

def add_class_ID(fulldataset, positive_class_label):
    labels = list(set(fulldataset['label_class']))
    # force positive class to be in position 1
    labels.remove(positive_class_label)
    labels.insert(1,positive_class_label)
    # assign class ID (integer with 1 being the positive class)
    IDs = [*range(0,len(labels))]
    fulldataset.insert(0, 'class_ID', -1)
    for n, label in enumerate(labels):
        mask = fulldataset['label_class'] == label
        fulldataset.loc[mask,'class_ID'] = IDs[n]
        class_encoder = pd.DataFrame({'label':labels, 'ID': IDs})
    return fulldataset, class_encoder

def add_subclass(fulldataset):
    # subclass for split = label_class + deployment_ID
    fulldataset.insert(0, 'subclass_label', fulldataset['label_class'] + '__' + fulldataset['deployment_ID'])
    labels = list(set(fulldataset['subclass_label']))
    IDs = [*range(0,len(labels))]
    fulldataset.insert(0,'subclass_ID', -1)
    for n, label in enumerate(labels):
        fulldataset.loc[fulldataset['subclass_label'] == label, 'subclass_ID'] = IDs[n]
    class_encoder = pd.DataFrame({'label':labels, 'ID': IDs})
    return fulldataset, class_encoder

def subclass2class_conversion(fulldataset):
    subclass_labels = list(set(fulldataset['subclass_label']))
    subclass_IDs = [*range(0,len(subclass_labels))]
    class_labels = []
    class_IDs = []
    for n, subclass_label in enumerate(subclass_labels):
        idx = fulldataset.index[fulldataset['subclass_label'] == subclass_labels[n]].tolist()
        class_labels.append(fulldataset.iloc[idx[0]]['label_class'])
        class_IDs.append(fulldataset.iloc[idx[0]]['class_ID'])
    class_encoder = pd.DataFrame({'subclass_labels': subclass_labels, 'subclass_ID': subclass_IDs, 'class_labels': class_labels, 'class_IDs': class_IDs})
    return class_encoder

def add_group(fulldataset):
    # # groups for splits = label_class + dateHour + deployment_ID
    fulldataset.insert(0,'TimeLabel',fulldataset['time_min_date'].dt.round("H").apply(lambda x: x.strftime('%Y%m%d%H%M%S')))
    # subclass for split = label_class + deployment_ID
    fulldataset.insert(0,'group_label',fulldataset['label_class'] + '_' + fulldataset['TimeLabel'] + '_' + fulldataset['deployment_ID'])
    labels = list(set(fulldataset['group_label']))
    IDs = [*range(0,len(labels))]
    fulldataset.insert(0,'group_ID', -1)
    for n, label in enumerate(labels):
        fulldataset.loc[fulldataset['group_label'] == label, 'group_ID'] = IDs[n]
    encoder = pd.DataFrame({'label':labels, 'ID': IDs})
    fulldataset.drop(columns = ['TimeLabel'])
    return fulldataset, encoder

def plot_dataset_distrib(dataset,attr_list=['subclass_label'], title=None):
    title = title + ' (' + str(len(dataset)) + ' data points)'
    nb_plots=len(attr_list)
    fig, ax = plt.subplots(1, nb_plots,
                           sharey=False,
                           constrained_layout=True,)
    for i in range(0,nb_plots):
        if nb_plots == 1:
            current_ax = ax
        else:
            current_ax = ax[i]
        #distrib = data_train.groupby(attr_list[i])[attr_list[i]].count().to_frame()
        distrib = dataset.groupby(attr_list[i])[attr_list[i]].count().to_frame()
        distrib['pct']= distrib[attr_list[i]]/ sum(distrib[attr_list[i]])*100
        #current_ax.bar(distrib.index,distrib['pct'], color='bkrgymc')
        current_ax.bar(distrib.index,distrib['pct'])
        current_ax.set_ylabel('Distribution of data points (%)')
        current_ax.set_title(attr_list[i])
        current_ax.grid()
        current_ax.tick_params(labelrotation=90 )
        #plt.xticks()
    fig.suptitle(title, fontsize=12)

def plot_datasets_groups(data_train, data_test, show=True):
    train_groups = list(set(data_train['group_ID']))
    test_groups = list(set(data_test['group_ID']))
    groups_intersection = list(set(train_groups) & set(test_groups))
    if show:
        plt.figure()
        #plt.bar(['Train set','Test set'],[len(train_groups),len(test_groups)],color='bkrgymc')
        plt.bar(['Train set','Test set'],[len(train_groups),len(test_groups)])
        plt.ylabel('Number of unique groups')
        plt.grid()
        plt.title('Number of shared groups: ' + str(len(groups_intersection)))
    return groups_intersection

def plot_datasets_distrib(data_train, data_test):
    ntrain = len(data_train)
    ntest = len(data_test)
    ntotal = ntrain + ntest
    ntrain = (ntrain/ntotal)*100
    ntest = (ntest/ntotal)*100
    plt.figure()
    #plt.bar(['Train set','Test set'],[ntrain,ntest], color='bkrgymc')
    plt.bar(['Train set','Test set'],[ntrain,ntest])
    plt.ylabel('% of data points')
    plt.title('Train/test sets data repartition')
    plt.grid()

def calc_tp_fp_fn_tn(Y_true,Y_prob,threshold):

    # init
    tp = np.zeros(len(Y_prob))
    fp = np.zeros(len(Y_prob))
    fn = np.zeros(len(Y_prob))
    tn = np.zeros(len(Y_prob))
    # thresholding
    Y_pred = np.zeros(len(Y_prob))
    Y_pred[Y_prob>=threshold] = 1
    idx=-1
    for true, pred in zip(Y_true, Y_pred):
        idx+=1
        if (true == 1) & (pred == 1): # true positive
            tp[idx]=1
        elif (true == 0) & (pred == 1): # false positive
            fp[idx]=1
        elif (true == 1) & (pred == 0): # false negative
            fn[idx]=1
        elif (true == 0) & (pred == 0): # true negative
            tn[idx]=1
    return tp, fp, fn, tn

def calc_performance_metrics(Y_true,Y_prob,thresholds=np.arange(0,1.01,0.01)):
    n = len(thresholds)
    precision = np.zeros(n)
    recall = np.zeros(n)
    f1_score = np.zeros(n)
    AUC_PR=0
    AUC_f1=0
    for idx, threshold in enumerate(thresholds):
        tp, fp, fn, tn = calc_tp_fp_fn_tn(Y_true,Y_prob,threshold=threshold)
        tp_tot = sum(tp)
        fp_tot = sum(fp)
        fn_tot = sum(fn)
        if (tp_tot == 0) | (fp_tot == 0):
            precision[idx] = np.nan
        else:
            precision[idx] = tp_tot /(tp_tot + fp_tot)
        recall[idx] = tp_tot /(tp_tot + fn_tot)
        f1_score[idx] = (2*precision[idx]*recall[idx]) / (precision[idx]+recall[idx])

    AUC_PR = metrics.auc(recall, precision)
    AUC_f1 = metrics.auc(thresholds, f1_score)
    out = pd.DataFrame({'thresholds': thresholds,'precision':precision,'recall':recall,'f1-score':f1_score})
    #out['AUC-PR'] = AUC_PR
    #out['AUC-f1'] = AUC_f1
    return out

def cross_validation(data_train, models, features, cv_splits=10,cv_repeats=10, rebalance=True):
    cv_predictions = pd.DataFrame({'CV_iter':[],'classifier':[],'uuid':[],'Y_true':[],'Y_pred':[],'Y_prob':[]})
    cv_performance = pd.DataFrame({'CV_iter':[],'classifier':[],'precision':[],'recall':[],'f1-score':[],'thresholds':[]})
    skf = RepeatedStratifiedGroupKFold(n_splits=cv_splits, n_repeats=cv_repeats, random_state=None)
    it=-1
    for cv_train_index, cv_val_index in skf.split(data_train, data_train['subclass_ID'],groups=data_train['group_ID']):
        it+=1
        # Split data train vs validation
        cv_data_train, cv_data_val = data_train.iloc[cv_train_index], data_train.iloc[cv_val_index]
        groups_intersection = plot_datasets_groups(cv_data_train, cv_data_val, show=False)
        # CV summary counts
        distrib_train = cv_data_train.groupby('label_class')['label_class'].count().to_frame()
        distrib_train.rename(columns={'label_class':'train'}, inplace=True)
        distrib_val = cv_data_val.groupby('label_class')['label_class'].count().to_frame()
        distrib_val.rename(columns={'label_class':'Validation'}, inplace=True)
        cv_summary = pd.concat([distrib_train, distrib_val],axis=1)
        # display CV info
        print(' ')
        print(' ')
        print('Cross validation #', str(it) + ' ---------------------------------------')
        print(cv_summary)
        print('Intersecting groups:' + str(len(groups_intersection)))

        #plot_dataset_distrib(cv_data_train,attr_list=['subclass_label','label_class'],title='Training set (CV #' + str(it) +')' )
        #plot_dataset_distrib(cv_data_val,attr_list=['subclass_label','label_class'],title='Evaluation set (CV #' + str(it) +')' )
        # reformat data
        X_train = cv_data_train[features] # features
        Y_train = cv_data_train['class_ID'] #labels
        X_val = cv_data_val[features] # features
        Y_val = cv_data_val['class_ID'] #labels
        Y_uuid = cv_data_val['uuid']

        # feature normalization
        Norm_mean = X_train.mean()
        Norn_std = X_train.std()
        X_train = (X_train-Norm_mean)/Norn_std
        X_val = (X_val-Norm_mean)/Norn_std

        print('  Positive training samples:', sum(Y_train == 1))
        print('  Negative training samples:', sum(Y_train == 0))

        if rebalance:
            # Making balanced dataset by oversampling
            print('Resampling training set with SMOTE + TomekLinks')
            #oversample = SMOTE(sampling_strategy='minority')
            #X_train, Y_train = oversample.fit_resample(X_train, Y_train)
            #undersample = RandomUnderSampler(sampling_strategy=0.5)
            #X_train, Y_train = undersample.fit_resample(X_train, Y_train)
            resample = SMOTETomek(tomek=TomekLinks(sampling_strategy='majority'))
            X_train, Y_train = resample.fit_resample(X_train, Y_train)
            print('  Positive samples:', sum(Y_train == 1))
            print('  Negatice samples:', sum(Y_train == 0))

        # Train and predict
        print('Classifiers:')
        for model_name, model in models:
            print('-> ' + model_name)
            # train model
            model.fit(X_train, Y_train)
            # predict
            pred_class = model.predict(X_val)
            pred_prob = model.predict_proba(X_val)
            # stack prediction info
            tmp = pd.DataFrame({'CV_iter':[],'classifier':[],'uuid':[],'Y_true':[],'Y_pred':[],'Y_prob':[]})
            tmp['uuid']= cv_data_val['uuid']
            tmp['CV_iter'] = it
            tmp['classifier'] = model_name
            tmp['Y_true'] = Y_val
            tmp['Y_pred'] = pred_class
            tmp['Y_prob'] = pred_prob[:,1]
            cv_predictions = pd.concat([cv_predictions,tmp],ignore_index=True)
            # calculate performance metrics
            performance = calc_performance_metrics(Y_val.values,pred_prob[:,1])
            performance['classifier'] = model_name
            performance['CV_iter'] = it
            cv_performance = pd.concat([cv_performance,performance],ignore_index=True)
    return cv_predictions, cv_performance

def summarize_performance(cv_performance, threshold=0.5):
    # evaluate predictions
    summary = pd.DataFrame({'Classifier':[],'Precision (mean)':[],'Precision (std)':[],'Recall (mean)':[],'Recall (std)':[],'f1-score (mean)':[],'f1-score (std)':[]})
    # plot PR curves
    classifiers = list(set(cv_performance['classifier']))
    cv_iterations = list(set(cv_performance['CV_iter']))
    for classifier in classifiers:
        temp_classif = cv_performance[cv_performance['classifier']==classifier]
        temp_classif = temp_classif[temp_classif['thresholds']==threshold]
        p_mean = round(temp_classif['precision'].mean(),3)
        p_std = round(temp_classif['precision'].std(),3)
        r_mean = round(temp_classif['recall'].mean(),3)
        r_std = round(temp_classif['recall'].std(),3)
        f_mean = round(temp_classif['f1-score'].mean(),3)
        f_std = round(temp_classif['f1-score'].std(),3)
        tmp = pd.DataFrame({'Classifier': [classifier],'Precision (mean)': [p_mean],'Precision (std)':[p_std],'Recall (mean)':[r_mean],'Recall (std)':[r_std],'f1-score (mean)':[f_mean],'f1-score (std)':[f_std]})
        summary = pd.concat([summary, tmp], ignore_index=True)
    return summary.T

def plot_PR_curves(cv_performance):
    # plot PR curves
    classifiers = list(set(cv_performance['classifier']))
    fig, ax = plt.subplots(1, 1,
                               sharey=False,
                               constrained_layout=True,)
    for classifier in classifiers:
        temp = cv_performance[cv_performance['classifier']==classifier]
        temp2 = temp.groupby(['thresholds']).mean()
        ax.plot(temp2['recall'],temp2['precision'], label=classifier)
    ax.set_ylabel('Precision')
    ax.set_xlabel('Recall')
    ax.set_title('Average Precision and Recall curve')
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.grid()
    ax.legend()

def plot_F_curves(cv_performance):
    # plot PR curves
    classifiers = list(set(cv_performance['classifier']))
    fig, ax = plt.subplots(1, 1,
                               sharey=False,
                               constrained_layout=True,)
    for classifier in classifiers:
        temp = cv_performance[cv_performance['classifier']==classifier]
        temp2 = temp.groupby(['thresholds']).mean()
        ax.plot(temp2.index,temp2['f1-score'], label=classifier)
    ax.set_ylabel('f1-score')
    ax.set_xlabel('Decision threshold')
    ax.set_title('Average f1-score curve')
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.grid()
    ax.legend()

def classification_train(X_train, Y_train, model, rebalance=True):
    if rebalance:
        # Making balanced dataset by oversampling
        print('Resampling training set with SMOTE + TomekLinks')
        resample = SMOTETomek(tomek=TomekLinks(sampling_strategy='majority'))
        X_train, Y_train = resample.fit_resample(X_train, Y_train)
        print('  Positive training samples:', sum(Y_train == 1))
        print('  Negative training samples:', sum(Y_train == 0))
    model_trained = model.fit(X_train, Y_train)
    return model_trained

def classification_predict(X_test, model_trained):
    pred_class = model_trained.predict(X_test)
    pred_prob = model_trained.predict_proba(X_test)
    return pred_class, pred_prob[:,1]

def plot_2d_space(X, y, label='Classes'):
    plt.figure()
    colors = ['#1F77B4', '#FF7F0E']
    markers = ['o', '.']
    for l, c, m in zip(np.unique(y), colors, markers):
        plt.scatter(
            X[y==l, 0],
            X[y==l, 1],
            c=c, label=l, alpha = 0.5, marker=m
        )
    plt.title(label)
    plt.legend(loc='upper right')
    plt.show()

def main():
    # input arguments
    input_args = dict()
    input_args['positive_class_label'] ='FS'
    input_args['train_ratio'] = 0.75
    input_args['cv_splits'] = 10 #5
    input_args['cv_repeats'] = 1
    input_args['rebalance_classes'] = True
    #input_args['data_file']= r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\results\dataset_FS-NN_modified_20201105145300.nc'
    input_args['data_file']= r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\results\dataset_FS-NN_modified_20200902194334.nc'
    input_args['out_dir'] = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\results\Classification'
    input_args['run_CV'] = False
    input_args['train_final_model'] = True
    input_args['final_model_name'] = 'RF50'

    ## DEFINITION OF CLASSIFIERS -------------------------------------------------
    models = []
    models.append(('Dummy', DummyClassifier(strategy="constant",constant=1)))
    models.append(('LR', LogisticRegression(solver='liblinear', multi_class='ovr')))
    models.append(('LDA', LinearDiscriminantAnalysis()))
    #models.append(('KNN', KNeighborsClassifier()))
    #models.append(('KNN', KNeighborsClassifier(n_neighbors=4, metric='euclidean')))
    models.append(('CART', DecisionTreeClassifier()))
    #models.append(('NB', GaussianNB()))
    models.append(('XGBoost', XGBClassifier()))
    #models.append(('MLP', MLPClassifier(solver='sgd', alpha=1e-5, hidden_layer_sizes=(5, 2), random_state=0)))
    models.append(('RF5', RandomForestClassifier(n_estimators=5,min_samples_split= 100, min_samples_leaf=50,random_state=0)))
    models.append(('RF10', RandomForestClassifier(n_estimators=10,min_samples_split= 100, min_samples_leaf=50,random_state=0)))
    models.append(('RF30', RandomForestClassifier(n_estimators=30,min_samples_split= 100, min_samples_leaf=50, random_state=0)))
    models.append(('RF50', RandomForestClassifier(n_estimators=50,min_samples_split= 100, min_samples_leaf=50, random_state=0)))
    #models.append(('RF100', RandomForestClassifier(n_estimators=100,min_samples_split= 100, min_samples_leaf=50,random_state=0)))

    ## setup output folder
    now = datetime.now()
    now_str = now.strftime("%Y%m%dT%H%M%S")
    out_dir = os.path.join(input_args['out_dir'],now_str)
    os.mkdir(out_dir)

    ## Save input args to txt file
    text_file = open(os.path.join(out_dir, 'input_args_' + now_str + '.txt'), "w")
    n = text_file.write(str(input_args))
    text_file.close()

    ## Checks that model name exists before running all the processing
    if input_args['train_final_model']:
        model_idx = [model[0] for model in models].index(input_args['final_model_name'] )

    ## LOAD DATSET ---------------------------------------------------------------
    dataset = Measurement()
    dataset.from_netcdf(input_args['data_file'])
    print(dataset.summary())

    ## DATA PREPARATION ----------------------------------------------------------
    # features
    features = dataset.metadata['measurements_name'][0] # list of features used for the classification
    # data
    data = dataset.data
    # drop FS observations at Mill Bay
    indexNames = data[(data['label_class'] == 'FS') & (data['location_name'] == 'Mill bay') ].index
    data.drop(indexNames, inplace=True)
    # add subclass + IDs
    data, class_encoder = add_class_ID(data, input_args['positive_class_label'])
    data, _ = add_subclass(data)
    #subclass2class_table = subclass2class_conversion(data)
    # add group ID
    data, group_encoder = add_group(data)

    ## DATA CLEAN-UP -------------------------------------------------------------
    # Basic stats on all features
    data_stats = data[features].describe()
    #print(data_stats)

    # how many NaNs and Infs per column
    data = data.replace([np.inf, -np.inf], np.nan)
    Nnan = data[features].isna().sum()
    ax = Nnan.plot(kind='bar',title='Number of NaN/Inf',grid=True)
    ax.set_ylabel('Number of observations with NaNs/Infs')

    # Drop some features with too many NaNs
    features.remove('freq_flatness')
    features.remove('snr')
    features.remove('uuid')

    # drop observations/rows with NaNs
    data.dropna(subset=features, axis=0, how='any', thresh=None, inplace=True)
    data_stats2 = data[features].describe()

    # ## VISUALIZATION -------------------------------------------------------------
    # # box and whisker plots
    # data[features].plot(kind='box', subplots=True, layout=(7,7), sharex=False, sharey=False)
    # # histograms
    # data[features].hist()
    # # scatter plot matrix
    # pd.plotting.scatter_matrix(data[features])
    # scatter plot PCA
    # pca = PCA(n_components=2)
    # X  = pca.fit_transform(data[features])
    # y = data['class_ID']
    # plot_2d_space(X, y, 'Imbalanced dataset (2 PCA components)')

    ## SPLIT DATA INTO TRAIN & TEST SETS ------------------------------------------
    n_splits = round(1/(1-input_args['train_ratio']))
    skf = StratifiedGroupKFold(n_splits=n_splits, shuffle=True, random_state=None)
    for train_index, test_index in skf.split(data, data['subclass_ID'],groups=data['group_ID']):
        data_train, data_test = data.iloc[train_index], data.iloc[test_index]
        break
    # plot class repartition
    plot_datasets_distrib(data_train, data_test)
    plot_dataset_distrib(data,attr_list=['subclass_label','label_class'],title='Full dataset')
    plot_dataset_distrib(data_train,attr_list=['subclass_label','label_class'],title='Training set')
    plot_dataset_distrib(data_test,attr_list=['subclass_label','label_class'],title='Test set')
    # verify groups are not used in both datasets
    groups_intersection = plot_datasets_groups(data_train, data_test, show=True)

    ## CROSS VALIDATION ON TRAIN SET ----------------------------------------------
    if input_args['run_CV']:
        # run train/test experiments
        cv_predictions, cv_performance = cross_validation(data_train, models, features, cv_splits=input_args['cv_splits'],cv_repeats=input_args['cv_repeats'], rebalance=input_args['rebalance_classes'])
        # display summary results
        performance_report = summarize_performance(cv_performance, threshold=0.5)
        print(performance_report)
        # plot mean Precision and Recall curves
        plot_PR_curves(cv_performance)
        plot_F_curves(cv_performance)
        # save results
        CV_results ={'cv_predictions': cv_predictions,
                     'cv_performance': cv_performance,
                     'models': models,
                     'input_args': input_args,
                     }
        pickle.dump(CV_results, open(os.path.join(out_dir, 'CV_' + now_str + '.sav'), 'wb'))

    ## FINAL EVALUATION ON TEST SET -----------------------------------------------
    if input_args['train_final_model']:

        print(' ')
        print('Final evaluation on test set:')
        print(' ')

        model_name = models[model_idx][0]
        model = models[model_idx][1] # RF50
        print(model)
        X_train = data_train[features] # features
        Y_train = data_train['class_ID'] #labels
        X_test = data_test[features] # features
        Y_test = data_test['class_ID'] #labels
        # feature normalization
        Norm_mean = X_train.mean()
        Norm_std = X_train.std()
        X_train = (X_train-Norm_mean)/Norm_std
        X_test = (X_test-Norm_mean)/Norm_std
        # Train on entire train set
        final_model = classification_train(X_train, Y_train, model, rebalance=input_args['rebalance_classes'])
        # Evaluate on full test set
        pred_class, pred_prob = classification_predict(X_test, final_model)
        # Print evaluation report
        CR = classification_report(Y_test, pred_class)
        print(CR)
        # save the model to disk
        model= {'name': model_name,
                'model':final_model,
                'features': features,
                'normalization_mean': Norm_mean,
                'normalization_std': Norm_std,
                'classes': class_encoder,
                'input_args': input_args,
                }
        pickle.dump(model, open(os.path.join(out_dir, model_name + '_model_' + now_str + '.sav'), 'wb'))
        # precision, recall, thresholds = precision_recall_curve(Y_val, pred_prob[:,0])
        # pr_auc = metrics.auc(recall, precision)
        # f1 = f1_score(Y_val, pred_class, average='binary')
        # CR = classification_report(Y_val, pred_class)
        # CM = confusion_matrix(Y_val, pred_class)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
"""
Created on Mon May 25 14:56:39 2020

@author: xavier.mouy
"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import time
import pandas as pd

## Input paraneters ##########################################################

audio_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\audio_data\AMAR173.4.20190916T011248Z.wav"
annotation_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\manual_annotations\AMAR173.4.20190916T011248Z.Table.1.selections.txt"
detection_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset_with_metadata\AMAR173.4.20190916T011248Z.wav.nc"

# Spectrogram parameters
frame = 3000
nfft = 4096
step = 500
#ovlp = 2500
fmin = 0
fmax = 1000
window_type = 'hann'

# start and stop time of wavfile to analyze
t1 = 0#24
t2 = 200#40
## ###########################################################################
tic = time.perf_counter()

# load audio data
sound = Sound(audio_file)
sound.read(channel=0, chunk=[t1, t2], unit='sec', detrend=True)

# Calculates  spectrogram
spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='samp')
spectro.compute(sound, dB=True, use_dask=True, dask_chunks=40)
spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)

# load annotations
annot = Annotation()
annot.from_raven(annotation_file)

# load detections
detec = Measurement()
detec.from_netcdf(detection_file)

freq_ovp = True # default True
dur_factor_max = None # default None
dur_factor_min = 0.1 # default None
ovlp_ratio_min = 0.3 # defaulkt None
remove_duplicates = True # dfault - False
inherit_metadata = True # default False
filter_deploymentID = False # default True
# here <-= filter per filename and deployment ID

# stack = []
# det = detec.data
# for index, an in annot.data.iterrows(): #for each annotation
#     ## restrict to the specific deploymnetID of the annotation if file names are not unique
#     if filter_deploymentID:
#         df = det[det.deployment_ID == an.deployment_ID]
#     ## filter detections to same file and deployment ID as the current annotation
#     df = det[det.audio_file_name == an.audio_file_name]
#     ## check overlap in time first
#     if len(df) > 0:
#         df = det[((det.time_min_offset <= an.time_min_offset) & (det.time_max_offset >= an.time_max_offset)) |  # 1- annot inside detec
#                  ((det.time_min_offset >= an.time_min_offset) & (det.time_max_offset <= an.time_max_offset)) |  # 2- detec inside annot
#                  ((det.time_min_offset < an.time_min_offset) & (det.time_max_offset < an.time_max_offset) & (det.time_max_offset > an.time_min_offset)) | # 3- only the end of the detec overlaps with annot
#                  ((det.time_min_offset > an.time_min_offset) & (det.time_min_offset < an.time_max_offset) & (det.time_max_offset > an.time_max_offset)) # 4- only the begining of the detec overlaps with annot
#                   ]
#     # then looks at frequency overlap. Can be turned off if freq bounds are not reliable
#     if (len(df) > 0) & freq_ovp:
#         df = df[((df.frequency_min <= an.frequency_min) & (df.frequency_max >= an.frequency_max)) | # 1- annot inside detec
#                 ((df.frequency_min >= an.frequency_min) & (df.frequency_max <= an.frequency_max)) | # 2- detec inside annot
#                 ((df.frequency_min < an.frequency_min) & (df.frequency_max < an.frequency_max) & (df.frequency_max > an.frequency_min)) | # 3- only the top of the detec overlaps with annot
#                 ((df.frequency_min > an.frequency_min) & (df.frequency_min < an.frequency_max) & (df.frequency_max > an.frequency_max)) # 4- only the bottom of the detec overlaps with annot
#                 ]
#     # discard if durations are too different 
#     if (len(df) > 0) & (dur_factor_max is not None):
#         df = df[df.duration < an.duration*dur_factor_max]
#     if (len(df) > 0) & (dur_factor_min is not None):
#         df = df[df.duration > an.duration*dur_factor_min]

#     # discard if they don't overlap enough
#     if (len(df) > 0) & (ovlp_ratio_min is not None):
#         df_ovlp = (df['time_max_offset'].apply(lambda x: min(x,an.time_max_offset)) - df['time_min_offset'].apply(lambda x: max(x,an.time_min_offset))) / an.duration
#         df = df[df_ovlp>=ovlp_ratio_min]

#     if (len(df) > 1) & remove_duplicates:
#         df = df.iloc[[df_ovlp.values.argmax()]] # pick teh one with max time overlap
        
#     if len(df) > 0:
#         if inherit_metadata:
#             df['mooring_platform_name'] = an['mooring_platform_name']
#             df['recorder_type'] = an['recorder_type']
#             df['recorder_SN'] = an['recorder_SN']
#             df['hydrophone_model'] = an['hydrophone_model']
#             df['hydrophone_SN'] = an['hydrophone_SN']
#             df['hydrophone_depth'] = an['hydrophone_depth']
#             df['location_name'] = an['location_name']
#             df['location_lat'] = an['location_lat']
#             df['location_lon'] = an['location_lon']
#             df['location_water_depth'] = an['location_water_depth']
#             df['deployment_ID'] = an['deployment_ID']
#             df['label_class'] = an['label_class']
#             df['label_subclass'] = an['label_subclass']
#         stack.append(df)
# ovlp = pd.concat(stack, ignore_index=True)
# detec.data = ovlp
# print(len(detec))

detec.filter_overlap_with(annot, 
                          freq_ovp=freq_ovp,
                          dur_factor_max=dur_factor_max,
                          dur_factor_min=dur_factor_min,
                          ovlp_ratio_min=ovlp_ratio_min,
                          remove_duplicates=remove_duplicates,
                          inherit_metadata=inherit_metadata,
                          filter_deploymentID=filter_deploymentID,
                          inplace=True
                          )
print(len(detec))    

# # detec.overlap_with(annot)
# toc = time.perf_counter()
# print(f"Executed in {toc - tic:0.4f} seconds")

# Plot
graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000, time_min=20, time_max=200)
graph.add_data(sound)
graph.add_annotation(detec, panel=0, color='green', label='Detections')
graph.add_data(spectro)
graph.add_annotation(annot, panel=1, color='red', label='Annotations')
graph.add_annotation(detec, panel=1, color='green', label='Detections')
graph.colormap = 'binary'
#graph.colormap = 'jet'
graph.show()




# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 10:39:23 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import time
import pandas as pd

annotation_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\Master_annotations_dataset.nc"
detec_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\dataset_annotations_only.nc'

# # load annotations
# print('-----------------')
# print('  Annotations    ')
# annot = Annotation()
# annot.from_netcdf(annotation_file)
# print(annot.summary())
# annot_perfile = annot.summary(rows='audio_file_name',columns='label_class')
# annot_perfile.rename(columns={"FS": "FS-annot"}, inplace=True)
# annot_perfile = annot_perfile['FS-annot'].to_frame()
# #annot_perfile.to_csv('annot.csv')

print(' ')
print('-----------------')
print('  Detections     ')
# load detections
detec = Measurement()
detec.from_netcdf(detec_file)
print(detec.summary())
detec_perfile = detec.summary(rows='audio_file_name',columns='label_class')
detec_perfile.rename(columns={"FS": "FS-detec"}, inplace=True)
detec_perfile = detec_perfile['FS-detec'].to_frame()


dd= pd.concat([annot_perfile,detec_perfile], axis=1)
dd['diff'] = dd['FS-annot'] - dd['FS-detec']
dd.plot()


# outdir=r'C:\Users\xavier.mouy\Documents\Workspace\GitHub\ecosound\tests\detec_export'
# detec.to_pamlab(outdir, single_file=False)


# outdir=r'C:\Users\xavier.mouy\Documents\Workspace\GitHub\ecosound\tests\annot_export'
# annot.to_pamlab(outdir, single_file=False)




# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 22:35:31 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound, Filter

#single_channel_file = r"../ecosound/resources/AMAR173.4.20190916T061248Z.wav"
multi_channel_file = r"../ecosound/resources/671404070.190722162836.wav"


# load part of the file and plot
print('------------------------------')
#sig = Sound() # should return error
sig = Sound(multi_channel_file)
print(len(sig))
sig.read(channel=0, chunk=[10, 1000])
print('------------------------------')
print(len(sig))
print('start sample: ', sig.waveform_start_sample)
print('stop sample: ', sig.waveform_stop_sample)
print('duration:: ', sig.waveform_duration_sample)
print(len(sig))
sig.plot_waveform(newfig=True)

# extract a sinppet from the data
sig2 = sig.select_snippet([100,1000])
sig2.plot_waveform(newfig=True)
print('------------------------------')
print('start sample: ', sig2.waveform_start_sample)
print('stop sample: ', sig2.waveform_stop_sample)
print('duration:: ', sig2.waveform_duration_sample)
print(len(sig2))

# try filter
filter_type = ['bandpass', 'lowpass', 'highpass']
cutoff_frequencies = [100, 1000]
sig2.filter(filter_type[0 ], cutoff_frequencies, order=4)
# try agfain -> should retrun error
#sig2.filter(filter_type[0], cutoff_frequencies, order=4)
sig2.plot_waveform(newfig=True)
print('------------------------------')
print('start sample: ', sig2.waveform_start_sample)
print('stop sample: ', sig2.waveform_stop_sample)
print('duration: ', sig2.waveform_duration_sample)
print(len(sig2))

# # re-adjust seletec waveform based on energy
# energy_percentage = 99.9
# sig2.tighten_waveform_window(energy_percentage)
# sig2.plot_waveform(newfig=True)
# print('------------------------------')
# print('start sample: ', sig2.waveform_start_sample)
# print('stop sample: ', sig2.waveform_stop_sample)
# print('duration:: ', sig2.waveform_duration_sample)
# print(len(sig2))


# sig3 = Sound(multi_channel_file)
# sig3.read(channel=0, chunk=[sig2.waveform_start_sample, sig2.waveform_stop_sample])
# #sig3.read(channel=0, chunk=[134, 1009])
# sig3.plot_waveform(newfig=True)
# print('------------------------------')
# print('start sample: ', sig3.waveform_start_sample)
# print('stop sample: ', sig3.waveform_stop_sample)
# print('duration:: ', sig3.waveform_duration_sample)
# print(len(sig3))



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 15:28:29 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement


# # ## netcdf folder
# netcdf_files = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test'
# annot4 = Annotation()
# annot4.from_netcdf(netcdf_files, verbose=True)
# print(len(annot4))


# # ## Load netcdf measurmeent folder from folder
# netcdf_files = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Noise_dataset'
# outfile=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Noise_dataset\dataset_noise.nc'
# meas = Measurement()
# meas.from_netcdf(netcdf_files, verbose=True)
# print(len(meas))
# #meas.to_netcdf(outfile)

# # ## Load netcdf measurmeent folder from single file
netcdf_files = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset_with_metadata2\JASCOAMARHYDROPHONE742_20140913T115018.797Z.wav.nc'
meas = Measurement()
meas.from_netcdf(netcdf_files, verbose=True)
print(len(meas))

# # ## Load netcdf measurmeent folder from list of files
# netcdf_files = []
# netcdf_files.append(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test2\67391492.181017121114.wav.nc")
# netcdf_files.append(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test2\67391492.181017151114.wav.nc")
# netcdf_files.append(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test2\67391492.181017181114.wav.nc")
# meas = Measurement()
# meas.from_netcdf(netcdf_files, verbose=True)
# print(len(meas))





from sklearn.model_selection._split import _BaseKFold
from sklearn.model_selection._split import _RepeatedSplits
import numpy as np
from collections import defaultdict, Counter
from sklearn.utils import check_random_state

class StratifiedGroupKFold(_BaseKFold):
    """Stratified K-Folds iterator variant with non-overlapping groups.

    This cross-validation object is a variation of StratifiedKFold that returns
    stratified folds with non-overlapping groups. The folds are made by
    preserving the percentage of samples for each class.

    The same group will not appear in two different folds (the number of
    distinct groups has to be at least equal to the number of folds).

    The difference between GroupKFold and StratifiedGroupKFold is that
    the former attempts to create balanced folds such that the number of
    distinct groups is approximately the same in each fold, whereas
    StratifiedGroupKFold attempts to create folds which preserve the
    percentage of samples for each class.

    Read more in the :ref:`User Guide <cross_validation>`.

    Parameters
    ----------
    n_splits : int, default=5
        Number of folds. Must be at least 2.

    shuffle : bool, default=False
        Whether to shuffle each class's samples before splitting into batches.
        Note that the samples within each split will not be shuffled.

    random_state : int or RandomState instance, default=None
        When `shuffle` is True, `random_state` affects the ordering of the
        indices, which controls the randomness of each fold for each class.
        Otherwise, leave `random_state` as `None`.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.model_selection import StratifiedGroupKFold
    >>> X = np.ones((17, 2))
    >>> y = np.array([0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    >>> groups = np.array([1, 1, 2, 2, 3, 3, 3, 4, 5, 5, 5, 5, 6, 6, 7, 8, 8])
    >>> cv = StratifiedGroupKFold(n_splits=3)
    >>> for train_idxs, test_idxs in cv.split(X, y, groups):
    ...     print("TRAIN:", groups[train_idxs])
    ...     print("      ", y[train_idxs])
    ...     print(" TEST:", groups[test_idxs])
    ...     print("      ", y[test_idxs])
    TRAIN: [2 2 4 5 5 5 5 6 6 7]
           [1 1 1 0 0 0 0 0 0 0]
     TEST: [1 1 3 3 3 8 8]
           [0 0 1 1 1 0 0]
    TRAIN: [1 1 3 3 3 4 5 5 5 5 8 8]
           [0 0 1 1 1 1 0 0 0 0 0 0]
     TEST: [2 2 6 6 7]
           [1 1 0 0 0]
    TRAIN: [1 1 2 2 3 3 3 6 6 7 8 8]
           [0 0 1 1 1 1 1 0 0 0 0 0]
     TEST: [4 5 5 5 5]
           [1 0 0 0 0]

    See also
    --------
    StratifiedKFold: Takes class information into account to build folds which
        retain class distributions (for binary or multiclass classification
        tasks).

    GroupKFold: K-fold iterator variant with non-overlapping groups.
    """

    def __init__(self, n_splits=5, shuffle=False, random_state=None):
        super().__init__(n_splits=n_splits, shuffle=shuffle,
                         random_state=random_state)

    # Implementation based on this kaggle kernel:
    # https://www.kaggle.com/jakubwasikowski/stratified-group-k-fold-cross-validation
    def _iter_test_indices(self, X, y, groups):
        labels_num = np.max(y) + 1
        y_counts_per_group = defaultdict(lambda: np.zeros(labels_num))
        y_distr = Counter()
        for label, group in zip(y, groups):
            y_counts_per_group[group][label] += 1
            y_distr[label] += 1

        y_counts_per_fold = defaultdict(lambda: np.zeros(labels_num))
        groups_per_fold = defaultdict(set)

        groups_and_y_counts = list(y_counts_per_group.items())
        rng = check_random_state(self.random_state)
        if self.shuffle:
            rng.shuffle(groups_and_y_counts)

        for group, y_counts in sorted(groups_and_y_counts,
                                      key=lambda x: -np.std(x[1])):
            best_fold = None
            min_eval = None
            for i in range(self.n_splits):
                y_counts_per_fold[i] += y_counts
                std_per_label = []
                for label in range(labels_num):
                    std_per_label.append(np.std(
                        [y_counts_per_fold[j][label] / y_distr[label]
                         for j in range(self.n_splits)]))
                y_counts_per_fold[i] -= y_counts
                fold_eval = np.mean(std_per_label)
                if min_eval is None or fold_eval < min_eval:
                    min_eval = fold_eval
                    best_fold = i
            y_counts_per_fold[best_fold] += y_counts
            groups_per_fold[best_fold].add(group)

        for i in range(self.n_splits):
            test_indices = [idx for idx, group in enumerate(groups)
                            if group in groups_per_fold[i]]
            yield test_indices


class RepeatedStratifiedGroupKFold(_RepeatedSplits):
    """Repeated Stratified K-Fold cross validator.

    Repeats Stratified K-Fold with non-overlapping groups n times with
    different randomization in each repetition.

    Read more in the :ref:`User Guide <cross_validation>`.

    Parameters
    ----------
    n_splits : int, default=5
        Number of folds. Must be at least 2.

    n_repeats : int, default=10
        Number of times cross-validator needs to be repeated.

    random_state : int or RandomState instance, default=None
        Controls the generation of the random states for each repetition.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.model_selection import RepeatedStratifiedGroupKFold
    >>> X = np.ones((17, 2))
    >>> y = np.array([0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    >>> groups = np.array([1, 1, 2, 2, 3, 3, 3, 4, 5, 5, 5, 5, 6, 6, 7, 8, 8])
    >>> cv = RepeatedStratifiedGroupKFold(n_splits=2, n_repeats=2,
    ...                                   random_state=36851234)
    >>> for train_index, test_index in cv.split(X, y, groups):
    ...     print("TRAIN:", groups[train_idxs])
    ...     print("      ", y[train_idxs])
    ...     print(" TEST:", groups[test_idxs])
    ...     print("      ", y[test_idxs])
    TRAIN: [2 2 4 5 5 5 5 8 8]
           [1 1 1 0 0 0 0 0 0]
     TEST: [1 1 3 3 3 6 6 7]
           [0 0 1 1 1 0 0 0]
    TRAIN: [1 1 3 3 3 6 6 7]
           [0 0 1 1 1 0 0 0]
     TEST: [2 2 4 5 5 5 5 8 8]
           [1 1 1 0 0 0 0 0 0]
    TRAIN: [3 3 3 4 7 8 8]
           [1 1 1 1 0 0 0]
     TEST: [1 1 2 2 5 5 5 5 6 6]
           [0 0 1 1 0 0 0 0 0 0]
    TRAIN: [1 1 2 2 5 5 5 5 6 6]
           [0 0 1 1 0 0 0 0 0 0]
     TEST: [3 3 3 4 7 8 8]
           [1 1 1 1 0 0 0]

    Notes
    -----
    Randomized CV splitters may return different results for each call of
    split. You can make the results identical by setting `random_state`
    to an integer.

    See also
    --------
    RepeatedStratifiedKFold: Repeats Stratified K-Fold n times.
    """

    def __init__(self, n_splits=5, n_repeats=10, random_state=None):
        super().__init__(StratifiedGroupKFold, n_splits=n_splits,
                         n_repeats=n_repeats, random_state=random_state)


# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:41:54 2020

@author: xavier.mouy

"""
import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.detection.detector_builder import DetectorFactory
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import ecosound.core.tools
import time


## Input paraneters ##########################################################

#single_channel_file = r"../ecosound/resources/67674121.181018013806.wav"
single_channel_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\audio_data\AMAR173.4.20190916T004248Z.wav"
# Spectrogram parameters
frame = 3000
nfft = 4096
step = 500
#ovlp = 2500
fmin = 0
fmax = 1000
window_type = 'hann'

# start and stop time of wavfile to analyze
t1 = 1
t2 = 50
## ###########################################################################
tic = time.perf_counter()

# load audio data
sound = Sound(single_channel_file)
sound.read(channel=0, chunk=[t1, t2], unit='sec', detrend=True)

# Calculates  spectrogram
spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='samp')
spectro.compute(sound, dB=True, use_dask=True, dask_chunks=40)

# Crop unused frequencies
spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)

# Denoise
spectro.denoise('median_equalizer', window_duration=3, use_dask=True, dask_chunks=(2048,1000), inplace=True)

# Detector
detector = DetectorFactory('BlobDetector', kernel_duration=0.1, kernel_bandwidth=300, threshold=10, duration_min=0.05, bandwidth_min=40)
detections = detector.run(spectro, use_dask=True, dask_chunks=(2048,1000), debug=False)

# # Plot
# graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000)
# graph.add_data(sound)
# graph.add_annotation(detections, panel=0, color='red')
# graph.add_data(spectro)
# graph.add_annotation(detections, panel=1)
# #graph.colormap = 'binary'
# graph.colormap = 'jet'
# graph.show()

# Maasurements
spectro_features = MeasurerFactory('SpectrogramFeatures', resolution_time=0.01, resolution_freq=1, interp='linear')
measurements = spectro_features.compute(spectro,
                                        detections,
                                        debug=True,
                                        verbose=False,
                                        use_dask=True)
#measurements.to_netcdf('test.nc')

toc = time.perf_counter()
print(f"Executed in {toc - tic:0.4f} seconds")
# #########################################



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 16:43:47 2020

@author: xavier.mouy
"""

from .grapher_builder import BaseClass
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation

import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib
import numpy as np


class SoundPlotter(BaseClass):
    """A Grapher class to visualize sound data.

    Display wavforms, spectrograms of a sound signal and can also overlay
    annotations on each sound representation.

    The SoundPlotter grapher must be instantiated using the GrapherFactory with
    the positional argument 'SoundPlotter':

    from ecosound.visualization.grapher_builder import GrapherFactory
    graph = GrapherFactory('SoundPlotter',kwargs)

    Attributes
    ----------
    name : str
        Name of the grapher
    version : str
        Version of the grapher
    frequency_min : float
        Minimum frequency of the spectrograms, in Hz.
    frequency_max : float
        Maximum frequency of the spectrograms, in Hz.
    time_min : float
        Start time to graphs,in seconds.
    time_max : float
        End time to graphs,in seconds.
    unit : str
        Unit of the time axis. 'sec' or 'samp'
    fig_size : (float, float)
        Width, height of the figure, in inches.
    share_xaxis : bool
        Share x axis for all subplots.
    grid : bool
        Display grid in waveform plots.
    title : [str, str, ...] or str
        Title(s) of the figure/subplots.
    colormap : str
        Color map for the spectrogram. Uses names from Matplotlib. The default
        is 'jet'

    Methods
    -------
    add_data(*args)
        Add Sound or Spectrogram objects to plot.
    add_annotations(Annotation, panel=None, label=False, color='red')
        Add Annotation object to overlay on graphs.
    show(display=True)
        Display graph.
    to_file(filename)
        Save graph to file.
    """

    grapher_parameters = ('frequency_min',
                          'frequency_max',
                          'time_min',
                          'time_max',
                          'unit',
                          'fig_size',
                          'share_xaxis',
                          'grid',
                          'title',
                          'colormap',
                          )

    def __init__(self, *args, **kwargs):
        """
        Initialize graher object.

        Initialize the grapher object to display sound data.

        Parameters
        ----------
        *args : str
            Do not use. Only used by the GrapherFactory.
        frequency_min : float, optional
            Minimum frequency of the spectrograms, in Hz. The default is 0.
        frequency_max : float, optional
            Maximum frequency of the spectrograms, in Hz. The default is half
            of the Nyquist frequency.
        time_min : float, optional
            Start time to graphs,in seconds. The default is 0.
        time_max : float, optional
            End time to graphs,in seconds. The default is the end of the sound
            data.
        unit : str, optional
            Unit of the time axis: 'sec' for seconds; 'samp' for discrete (i.e.
            samples or bins for waveform and spectrogram, respectively). The
            default is 'sec'.
        fig_size : (float, float), optional
            Width, height of the figure, in inches. The default is (16, 4).
        share_xaxis : bool, optional
            Share x axis for all subplots. The default is True.
        grid : bool, optional
            Display grid in waveform plots. The default is True.
        title : [str, str, ...], str, None,  optional
            Title(s) of the figure/subplots. If a str, a single title is set
            for the entire figure at the top. If a list of str, each element i
            of the list defines the title for the subplot of index i. The
            number of elements in the list can equal or smaller than the number
            of subplots in the graph. A number of elements in teh list greater
            than the number of subplot will return an error. If set to None, no
            title will be displayed. The default is None.

        colormap : str
            Color map for the spectrogram. Uses names from Matplotlib (e.g.
            'jet', 'binary', 'gray', etc.). The default is 'jet'

        Returns
        -------
        None.

        """
        # Initialize all grapher parameters to None
        self.__dict__.update(dict(zip(self.grapher_parameters,
                                      [None]*len(self.grapher_parameters))))
        # Define default values:
        self.frequency_min = 0
        self.frequency_max = None
        self.time_min = 0
        self.time_max = None
        self.unit = 'sec'
        self.fig_size = (16, 4)
        self.share_xaxis = True
        self.grid = True
        self.title = None
        self.colormap = 'jet'
        # Unpack kwargs as grapher parameters if provided on instantiation
        self.__dict__.update(**kwargs)
        # Initialize containers
        self.data = []
        self.annotations = []

    @property
    def name(self):
        """Return name of the grapher."""
        grapher_name = 'SoundPlotter'
        return grapher_name

    @property
    def version(self):
        """Return version of the grapher."""
        version = '0.1'
        return version

    def add_data(self, *args):
        """
        Define sound or spectrogram data to plot.

        Add Sound or Spectrogram objects to plot. There is no restriction on
        the number of object to add. Each object will be displayed on a
        different subplot of the same figure. The order in which the objects
        are passed to this method defines the order in which they are displayed
        (i.e. subplots). When this method is called several times; objects are
        appended to the existing list of objects from previous calls.

        Parameters
        ----------
        *args : Spectrogram, Sound objects
            Sound and/or Spectrogram objects to display.

        Raises
        ------
        ValueError
            If no input arguments are provided.
            If the input arguments are not Sound or Spectrogram objects.

        Returns
        -------
        None. Updated grapher object.

        """
        if len(args) < 1:
            raise ValueError('There must be at least one input argument')
        # Check  type of each input arguments
        self._stack_data(args)

    def add_annotation(self, annotation, panel=None, label=False, color='red', tag=False, line_width=1):
        """
        Define annotations to display.

        When this method is called several times; ANnotation objects are
        appended to the existing list of objects from previous calls.

        Parameters
        ----------
        annotation : Annotation object.
            Annotation object to overlay on top of the waveforem or spectrogram
            plots.
        panel : int, list, None, optional
            Define on which subplot (i.e. data defined with the add_data
            method) to overlay the annotations on. If set to None, annotations
            will be displayed on all subplots. If set to int, annotations will
            only be displayed on the subplot index defined by panel. If set as
            a list of int, annotations will be displayed on the subplots
            indices defined in the panel panel. The default is None.
        label : bool, optional
            Display label for each annotation. Not implemented yet. The default
            is False.
        color : str, optional
            Color of the annotation boxes. Uses the color name as matplotlib
            (e.g. 'black', 'white','red', 'yellow', etc). The default is 'red'.
        tag : bool, optional
            If set to True, displays the classification confidence over each
            annotation box. The default is False.
		line_width : int, optional
            Width of the annotation line. The default is 1.

        Raises
        ------
        ValueError
            If the input argument is not an Annotation object.

        Returns
        -------
        None. Updated grapher object.

        """
        if isinstance(annotation, Annotation):
            self.annotations.append({'data': annotation,
                                     'panel': panel,
                                     'label': label,
                                     'color': color,
                                     'tag': tag,
									 'line_width': line_width,
                                     })
        else:
            raise ValueError('Type of input argument not recognized.'
                             'Accepted object type: Annotation')

    def show(self, display=True):
        """
        Display graph on screen.

        Display graph made of vertical subplots for each of the Sound and
        Spectrogram objects defined by the add_data method. The order in which
        the objects have been defined by add_data defines the order in which
        subplot they are displayed. Waveform and Spectrogram objects are
        identified automatically and plotted accordingly. Annotations are
        overlayed on subplots (i.e. panels) defined by the method
        add_annotations. Annotations are displayed as boxes in spectrograms,
        and as shaded areas in waveforms.

        Parameters
        ----------
        display : bool, optional
            Display figure on screen. The default is True.

        Raises
        ------
        ValueError
            If no data have been defined (see add_data method).
            If a Spectrogram has no data (i.e. spectrogram.compute() not excuted).
            If unit is not equal to 'sec' or 'samp'.

        Returns
        -------
        fig : Figure instance
            Figure object.
        ax : Axis instance
            Axis object with all subplots.

        """
        if len(self.data) == 0:
            raise ValueError('No data to plot. Use method .add_data to define'
                             ' the data to plot')
        # Display plot on screen?
        if display:
            matplotlib.use('Qt5Agg')
        else:
            matplotlib.use('Agg')
        # Define new figure and subplots
        nb_plots = len(self.data)
        fig, ax = plt.subplots(nb_plots, 1,
                               figsize=self.fig_size,
                               sharex=self.share_xaxis,
                               constrained_layout=True,
                               )  # gridspec_kw={'hspace': self.hspace}
        # Subplot titles
        titles = [None]*nb_plots
        if self.title is None:  # no titles
            pass
        if type(self.title) is str:
            titles[0] = self.title
        if type(self.title) is list:
            if len(self.title) > nb_plots:
                raise ValueError("More titles than subplots")
            else:
                titles[0:len(self.title)-1] = self.title
        # Plot data
        for idx, data in enumerate(self.data):
            if nb_plots == 1:
                current_ax = ax
            else:
                current_ax = ax[idx]
            if data['type'] == 'waveform':
                self._plot_waveform(data['data'], current_ax, title=titles[idx])
            elif data['type'] == 'spectrogram':
                self._plot_spectrogram(data['data'], current_ax, title=titles[idx])
            # only dipslay x label of bottom plot if shared axes
            if self.share_xaxis and (idx != nb_plots-1):
                current_ax.set_xlabel('')

        # Plot annotations
        for idx_annot, annot in enumerate(self.annotations):  # for each set of annotations
            # display annotations on all panels
            if annot['panel'] is None:
                annot['panel'] = range(0, nb_plots)
            # Make panel idx a list if not already
            if (type(annot['panel']) is float) or (type(annot['panel']) is int):
                annot['panel'] = [annot['panel']]
            # Check panel indices
            if max(annot['panel']) > nb_plots-1:
                raise ValueError("Invalid panel index")
            # PLot annotations on appropriate panels
            for idx_panel in annot['panel']:  # for each panel
                if len(self.data)==1:
                    current_ax = ax
                else:
                    current_ax = ax[idx_panel]
                self._plot_annotations(annot['data'], current_ax,
                                       panel_idx=idx_panel,
                                       label=annot['label'],
                                       color=annot['color'],
									   line_width = annot['line_width'],
                                       )
                if annot['label'] is not False:
                    handles, labels = current_ax.get_legend_handles_labels()
                    unique_labels=list(set(labels))
                    new_handles=[]
                    for l in unique_labels:
                        new_handles.append(handles[labels.index(l)])                        
                    current_ax.legend(new_handles,unique_labels,loc='upper right')

                if annot['tag'] is True:
                    bbox_props = dict(boxstyle="square", fc="w", ec="w", alpha=0.8)
                    panel_type = self.data[annot['panel'][0]]['type']
                    for index, row in annot['data'].data.iterrows():
                        if self.unit == 'sec':
                            #height = row['frequency_max']-row['frequency_min']
                            x = row['time_min_offset']
                            if panel_type == 'spectrogram':
                                y = row['frequency_max']
                            elif panel_type == 'waveform':
                                y = max(current_ax.get_ylim())
                            conf = str(round(row['confidence'],2))
                        elif self.unit == 'samp':
                            x = row['time_min_offset']
                            if panel_type == 'spectrogram':
                                y = row['frequency_max']
                            elif panel_type == 'waveform':
                                y = max(current_ax.get_ylim())
                            conf = str(round(row['confidence'],2))
                        current_ax.text(x, y, conf, size=8, bbox=bbox_props)
                    
                
        return fig, ax

    def to_file(self, filename):
        """
        Save the figure to an image file.

        Parameters
        ----------
        filename : str
            Full path of the image file to write.

        Returns
        -------
        None.

        """
        fig, _ = self.show(display=False)
        fig.savefig(filename, transparent=False, bbox_inches='tight',)

    def _plot_annotations(self, annot, ax, label, panel_idx, color, line_width):
        """Plot annotations on top of the waveform or spectrogram axes."""
        panel_type = self.data[panel_idx]['type']
        for index, row in annot.data.iterrows():
            # plot annotations on spectrograms
            if panel_type == 'spectrogram':
                alpha = 1
                facecolor = 'none'
                if self.unit == 'sec':
                    x = row['time_min_offset']
                    y = row['frequency_min']
                    width = row['duration']
                    height = row['frequency_max']-row['frequency_min']
                elif self.unit == 'samp':
                    time_resolution = self.data[panel_idx]['data'].time_resolution
                    x = round(row['time_min_offset']/time_resolution)
                    y = row['frequency_min']
                    width = round(row['duration']/time_resolution)
                    height = row['frequency_max']-row['frequency_min']
            elif panel_type == 'waveform':
                alpha = 0.2
                facecolor = color
                if self.unit == 'sec':
                    x = row['time_min_offset']
                    y = min(ax.get_ylim())
                    width = row['duration']
                    height = max(ax.get_ylim()) - min(ax.get_ylim())
                elif self.unit == 'samp':
                    time_resolution = self.data[panel_idx]['data'].waveform_sampling_frequency
                    x = round(row['time_min_offset']*time_resolution)
                    y = min(ax.get_ylim())
                    width = round(row['duration']*time_resolution)
                    height = max(ax.get_ylim()) - min(ax.get_ylim())
            rect = plt.Rectangle((x, y), width, height,
                                 linewidth=line_width,
                                 edgecolor=color,
                                 facecolor=facecolor,
                                 alpha=alpha,
                                 label=label)
            ax.add_patch(rect)

    def _stack_data(self, args):
        """Stack data to be plotted."""
        for idx, arg in enumerate(args):
            if isinstance(arg, Sound):
                self.data.append({'data': arg, 'type': 'waveform'})
            elif isinstance(arg, Spectrogram):
                self.data.append({'data': arg, 'type': 'spectrogram'})
            else:
                raise ValueError('Type of input argument not recognized.'
                                 'Accepted object types: Spectrogram, Sound')

    def _plot_spectrogram(self, spectro, current_ax, title=None):
        """Plot spectrogram on the current axis"""
        if self.frequency_max is None:
            self.frequency_max = spectro.sampling_frequency/2
        assert len(spectro.spectrogram) > 0, "Spectrogram not computed yet. "
        "Use the .compute() method first."
        if self.unit == 'sec':
            if self.time_max is None:
                self.time_max = spectro.axis_times[-1]
            current_ax.pcolormesh(spectro.axis_times,
                                  spectro.axis_frequencies,
                                  spectro.spectrogram,
                                  cmap=self.colormap,
                                  vmin=np.percentile(spectro.spectrogram, 50),
                                  vmax=np.percentile(spectro.spectrogram, 99.9),
								  shading='nearest',
                                  )
            xlabel = 'Time (sec)'
        elif self.unit == 'samp':
            axis_t = np.arange(0, len(spectro.axis_times), 1)
            if self.time_max is None:
                self.time_max = axis_t[-1]
            current_ax.pcolormesh(axis_t,
                                  spectro.axis_frequencies,
                                  spectro.spectrogram,
                                  cmap=self.colormap,
                                  vmin=np.percentile(spectro.spectrogram, 50),
                                  vmax=np.percentile(spectro.spectrogram, 99.9)
                                  )
            xlabel = 'Time (bin)'
        else:
            raise ValueError("Keyword 'unit' must be set to either 'sec' or"
                             " 'samp'.")
        current_ax.axis([self.time_min,
                         self.time_max,
                         self.frequency_min,
                         self.frequency_max]
                        )
        current_ax.set_ylabel('Frequency (Hz)')
        current_ax.set_xlabel(xlabel)
        current_ax.set_title(title)
        # if self.grid:
        #    current_ax.grid()
        return

    def _plot_waveform(self, sound, current_ax, title=None):
        """Plot waveform of a sound object on the current axis."""
        if len(sound._waveform) == 0:
            raise ValueError('Cannot plot, waveform data enpty. Use Sound.read'
                             + ' to load the waveform')
        if self.unit == 'sec':
            axis_t = np.arange(0, sound.waveform_duration_sample
                               / sound.waveform_sampling_frequency, 1
                               / sound.waveform_sampling_frequency)
            xlabel = 'Time (sec)'
        elif self.unit == 'samp':
            axis_t = np.arange(0, len(sound._waveform), 1)
            xlabel = 'Time (sample)'
        else:
            raise ValueError("Keyword 'unit' must be set to either 'sec' or"
                             " 'samp'.")
        if self.time_max is None:
                self.time_max = axis_t[-1]
        #axis_t = axis_t[0:len(sound._waveform)]
        current_ax.plot(axis_t, sound._waveform, color='black')
        current_ax.set_xlabel(xlabel)
        current_ax.set_ylabel('Amplitude')
        current_ax.set_title(title)
        
        current_ax.axis([self.time_min,
                         self.time_max,
                         min(sound._waveform),
                         max(sound._waveform)]
        # current_ax.axis([axis_t[0],
        #                  axis_t[-1],
        #                  min(sound._waveform),
        #                  max(sound._waveform)]
                        )
        if self.grid:
            current_ax.grid()



# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 21:06:18 2020

@author: xavier.mouy
"""
import xarray as xr
import os

indir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run2\RCA_in_April_July2019_1342218252'
file_ext = '.nc'
          
# loops thrugh each file
error_list =[]
files_list = os.listdir(indir) # list of files
for file in files_list:
    if os.path.isfile(os.path.join(indir, file)) & file.endswith(file_ext):
        try:
            ss = xr.open_dataset(os.path.join(indir,file))
        except:
            error_list.append(file)
                   


# -*- coding: utf-8 -*-
"""
Created on Sat May 30 10:19:38 2020

@author: xavier.mouy
"""


import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import time
import pandas as pd


## Input paraneters ##########################################################

annotation_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\Master_annotations_dataset.nc"
detection_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset_with_metadata2"
outfile=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\dataset_annotations_only2.nc'

# load annotations
annot = Annotation()
annot.from_netcdf(annotation_file)

# load detections
detec = Measurement()
detec.from_netcdf(detection_file)
print(detec)

freq_ovp = True # default True
dur_factor_max = None # default None
dur_factor_min = 0.1 # default None
ovlp_ratio_min = 0.3 # defaulkt None
remove_duplicates = True # dfault - False
inherit_metadata = True # default False
filter_deploymentID = False # default True

detec.filter_overlap_with(annot,
                          freq_ovp=freq_ovp,
                          dur_factor_max=dur_factor_max,
                          dur_factor_min=dur_factor_min,
                          ovlp_ratio_min=ovlp_ratio_min,
                          remove_duplicates=remove_duplicates,
                          inherit_metadata=inherit_metadata,
                          filter_deploymentID=filter_deploymentID,
                          inplace=True
                          )
print(detec)
detec.to_netcdf(outfile)


# -*- coding: utf-8 -*-
"""
Created on Fri May 15 14:07:09 2020

@author: xavier.mouy
"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.decorators import timeit

#from dask.distributed import Client, LocalCluster
import dask_image.ndfilters
import dask.array
import numpy as np
from scipy import ndimage
import dask
from numba import njit

@njit
def calcVariance2D(buffer):
    """Calculate the 2D variance."""
    return np.var(buffer)

@njit
def calcMedian(buffer):
    """Calculate the 2D variance."""
    return np.median(buffer)

@timeit
def Filter_med_normal(array, window_duration):
    # # Apply filter
    Smed = ndimage.median_filter(array, (1, window_duration))
    return Smed

@timeit
def Filter_med_numba(array, window_duration):
    # # Apply filter
    Svar = ndimage.generic_filter(array, calcMedian,
                                  (1, window_duration),
                                  mode='mirror')
    return Svar

@timeit
def Filter_var_normal(array,kernel_bandwidth, kernel_duration):
    # # Apply filter
    Svar = ndimage.generic_filter(array, calcVariance2D,
                                  (kernel_bandwidth, kernel_duration),
                                  mode='mirror')
    return Svar
@timeit
def Filter_var_dask(array,kernel_bandwidth, kernel_duration):
    dask_spectro = dask.array.from_array(array, chunks=(1000,1000))
    Svar = dask_image.ndfilters.generic_filter(dask_spectro,
                                               calcVariance2D,
                                               size=(kernel_bandwidth, kernel_duration),
                                               mode='mirror')
    Svar = Svar.compute()
    return Svar

@timeit
def Filter_med_dask(array, window_duration):
    dask_spectro = dask.array.from_array(array, chunks=(2048, 1000))
    Svar = dask_image.ndfilters.median_filter(dask_spectro,
                                               size=(1, window_duration),
                                               mode='mirror')
    Svar = Svar.compute()
    return Svar

# # Setup a local cluster.
# # By default this sets up 1 worker per core
# cluster = LocalCluster()
# client = Client(cluster)


array = np.random.rand(2048, 200000)

## Variance filter
# kernel_bandwidth = 100
# kernel_duration = 20
# S1 = Filter__var_normal(array,kernel_bandwidth, kernel_duration)
# S2 = Filter_var_dask(array, kernel_bandwidth, kernel_duration)

## Median filter
window_duration = 10
#Smed1 = Filter_med_normal(array, window_duration)
#Smed2 = Filter_med_numba(array, window_duration)
Smed3 = Filter_med_dask(array, window_duration)


# -*- coding: utf-8 -*-
"""
Created on Thu May 28 11:16:16 2020

@author: xavier.mouy
"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement
from ecosound.core.metadata import DeploymentInfo
import os
import ecosound.core.tools
import platform

outdir= r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset_with_metadata'
ext='.nc'


indir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset\UVIC_mill-bay_2019'
deployment_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_mill-bay_2019\deployment_info.csv'
data_dir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_mill-bay_2019\audio_data'

# load meta data
operator_name = platform.uname().node
dep_info = DeploymentInfo()
dep_info.read(deployment_file)

#list files
files = ecosound.core.tools.list_files(indir,
                                        ext,
                                        recursive=False,
                                        case_sensitive=True)

for idx,  file in enumerate(files):
    print(str(idx)+r'/'+str(len(files))+': '+ file)
    meas = Measurement()
    meas.from_netcdf(file)

    meas.insert_metadata(deployment_file)

    file_name = os.path.splitext(os.path.basename(file))[0]
    meas.insert_values(operator_name=platform.uname().node,
                       audio_file_name=os.path.splitext(os.path.basename(file_name))[0],
                       audio_file_dir=data_dir,
                       audio_file_extension='.wav',
                       audio_file_start_date= ecosound.core.tools.filename_to_datetime(file_name)[0]
                       )
    meas.to_netcdf(os.path.join(outdir,file_name+'.nc'))



# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:27:39 2020

@author: xavier.mouy
"""

from .measurer_builder import BaseClass
from ecosound.core.annotation import Annotation
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.measurement import Measurement
# from ecosound.visualization.grapher_builder import GrapherFactory
import ecosound.core.tools
import numpy as np
from scipy.stats import kurtosis, skew
from scipy.stats.mstats import gmean
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from numba import njit
from dask import delayed, compute, visualize

class SpectrogramFeatures(BaseClass):
    """Spectrogram features.

    This class extracts a set of spectral and temporal features from a
    spectrogram. It is based on measurements extracted by the software Raven
    and other measurements described in the R package Seewave by Jerome Sueur.

    The SpectrogramFeatures measurer must be instantiated using the
    MeasurerFactory with the positional argument 'SpectrogramFeatures':

    from ecosound.measurements.measurer_builder import MeasurerFactory
    spectro_features = MeasurerFactory('SpectrogramFeatures',
                                       resolution_time=0.001,
                                       resolution_freq=0.1,
                                       interp='linear')
    measurements = spectro_features.compute(spectro, detections,
                                            debug=False,
                                            verbose=False)

    The Measurement object returned has all the features appended to the
    original annotation fields in the pandas datafrane measurment.data.
    Measurer's name, version and features' name are in the pandas Dataframe
    measurement.metadata. Spectrogram features include:

        1- 'freq_peak': peak frequency in the frequency envelop, in Hz.
        2- 'freq_bandwidth': Bandwidth of the frequency envelop, in Hz.
        3- 'freq_bandwidth90': 90% bandwidth of the frequency envelop, in Hz.
        4- 'freq_pct5': frequency of the 5th percentile, in the frequency envelope, in Hz.
        5- 'freq_pct25': frequency of the 25th percentile, in the frequency envelope, in Hz.
        6- 'freq_pct50': frequency of the 50th percentile, in the frequency envelope, in Hz.
        7- 'freq_pct75': frequency of the 75th percentile, in the frequency envelope, in Hz.
        8- 'freq_pct95': frequency of the 95th percentile, in the frequency envelope, in Hz.
        9- 'freq_iqr': inter quartile range of the frequency envelope, in Hz.
       10- 'freq_asymmetry': symmetry of the frequency envelope.
       11- 'freq_concentration': concentration of the frequency envelope.
       12- 'freq_std': standard deciation of the frequency envelope.
       13- 'freq_kurtosis': kurtosis of the frequency envelope.
       14- 'freq_skewness': skewness of the frequency envelope.
       15- 'freq_entropy': Shannon's entropy of the frequency envelope.
       16- 'freq_flatness': flatness of the frequency envelope.
       17- 'freq_roughness': roughness of the frequency envelope.
       18- 'freq_centroid': centroid of the frequency envelope, in Hz.
       19- 'freq_overall_peak': overall peak frequency in the spectrogram, in Hz.
       20- 'freq_median_mean': mean of the median frequency through the spectrogram time slices, Hz.
       21- 'freq_median_std': standard deviation of the median frequency through the spectrogram time slices, Hz.
       22- 'freq_entropy_mean': mean of entropy through the spectrogram time slices, Hz.
       23- 'freq_entropy_std': std of the entropy through the spectrogram time slices, Hz.
       24- 'freq_upsweep_mean': frequency upsweep mean index
       25- 'freq_upsweep_fraction': frequency upsweep fraction
       26- 'snr': signal to noise ratio, in dB
       27- 'time_peak_sec': time of peak in the time envelope, in sec.
       28- 'time_peak_perc': relative time of peak in the time envelope.
       29- 'time_duration': duration of the time envelope, in sec.
       30- 'time_duration90': 90% duration of the time envelop, in sec.
       31- 'time_pct5': time of the 5th percentile, in the time envelope, in sec.
       32- 'time_pct25': time of the 25th percentile, in the time envelope, in sec.
       33- 'time_pct50': time of the 50th percentile, in the time envelope, in sec.
       34- 'time_pct75': time of the 75th percentile, in the time envelope, in sec.
       35- 'time_pct95': time of the 95th percentile, in the time envelope, in sec.
       36- 'time_iqr': inter quartile range of the time envelope, in sec.
       37- 'time_asymmetry': symmetry of the time envelope.
       38- 'time_concentration': concentration of the time envelope.
       39- 'time_std': standard deciation of the time envelope.
       40- 'time_kurtosis': kurtosis of the time envelope.
       41- 'time_skewness': skewness of the time envelope.
       42- 'time_entropy': Shannon's entropy of the time envelope.
       43- 'time_flatness': flatness of the time envelope.
       44- 'time_roughness': roughness of the time envelope.
       45- 'time_centroid': centroid of the time envelope, in sec.

    Attributes
    ----------
    name : str
        Name of the measurer
    version : str
        Version of the measurer
    resolution_freq : float
        frequency resolution of the interpolated spectral envelope, in Hz.
        Default is 0.1.
    resolution_time : float
        Time resolution of the interpolated temporal envelope, in seconds.
        Default is 0.001.
    interp : str
        Type of interpolation method for interpolating time and frequency
        envelopes. Can be 'linear' or 'quadratic'. Default is 'linear'.

    Methods
    -------
    run(spectrogram, detections, debug=True, verbose=False debug=False)
        Calculate spectrogram features each detection in the spectrogram.

    """

    measurer_parameters = ('resolution_freq',
                           'resolution_time',
                           'interp',
                           )

    def __init__(self, *args, **kwargs):
        """
        Initialize the measurer.

        Parameters
        ----------
        *args : str
            Do not use. Only used by the MeasurerFactory.
        resolution_freq : float, optional
            frequency resolution of the interpolated spectral envelope, in Hz.
            Default is 0.1.
        resolution_time : float, optional
            Time resolution of the interpolated temporal envelope, in seconds.
            Default is 0.001.
        interp : str, optional
            Type of interpolation method for interpolating time and frequency
            envelopes. Can be 'linear' or 'quadratic'. Default is 'linear'.

        Returns
        -------
        None. Measurer object.

        """
        # Initialize all measurer parameters to None
        self.__dict__.update(dict(zip(self.measurer_parameters,
                                      [None]*len(self.measurer_parameters))))
        # default values:
        self.resolution_time = 0.001
        self.resolution_freq = 0.1
        self.interp = 'linear'
        # Unpack kwargs as measurer parameters if provided on instantiation
        self.__dict__.update(**kwargs)

    @property
    def name(self):
        """Return name of the measurer."""
        measurer_name = 'SpectrogramFeatures'
        return measurer_name

    @property
    def version(self):
        """Return version of the measurer."""
        version = '0.1'
        return version

    def _prerun_check(self, spectrogram, annotations):
        """Run several verifications before the run."""
        # check that all required arguments are defined
        if True in [self.__dict__.get(keys) is None for keys in self.measurer_parameters]:
            raise ValueError('Not all measurer parameters have been defined.'
                             + ' Required parameters: '
                             + str(self.measurer_parameters))
        # check that spectrogram is a spectrogram class
        if not isinstance(spectrogram, Spectrogram):
            raise ValueError('Input must be an ecosound Spectrogram object'
                             + '(ecosound.core.spectrogram).')
        # check that annotations is an Annotation class
        if not isinstance(annotations, Annotation):
            raise ValueError('Input must be an ecosound Annotation object'
                             + '(ecosound.core.annotation).')

    def compute_old(self, spectro, annotations, debug=False, verbose=False):
        """ Compute spectrogram features.

        Goes through each annotation and compute features from the spectrogram.

        Parameters
        ----------
        spectro : ecosound Spectrogram object
            Spectrogram of the recording to analyze.
        annotations : ecosound Annotation object
            Annotations of the sounds to measure. Can be from manual analysis
            or from an automatic detector.
        debug : bool, optional
            Displays figures for each annotation with the spectrogram, spectral
            and time envelopes, and tables with all associated measurements.
            The default is False.
        verbose : bool, optional
            Prints in the console the annotation being processed. The default
            is False.

        Returns
        -------
        measurements : ecosound Measurement object
            Measurement object containing the measurements appended to the
            original annotation fields. Measurements are in the .data data
            frame. Metadata with mearurer name, version and measurements names
            are in the .metadata datafreame.

        """
        self._prerun_check(spectro, annotations)
        # loop through each annotation
        for index, annot in annotations.data.iterrows():
            if verbose:
                print('processing annotation ', index)
            tmin = annot['time_min_offset']
            tmax = annot['time_max_offset']
            fmin = annot['frequency_min']
            fmax = annot['frequency_max']
            # extract minmgram for that detection
            minigram = spectro.crop(frequency_min=fmin,
                                    frequency_max=fmax,
                                    time_min=tmin,
                                    time_max=tmax)
            # extract time and frequency envelops
            envelop_time, envelop_freq = SpectrogramFeatures.get_envelops(minigram,
                                                            normalize=True)
            # interpolate each envelop
            axis_t, envelop_time2 = ecosound.core.tools.resample_1D_array(
                minigram.axis_times,
                envelop_time,
                resolution=self.resolution_time,
                kind=self.interp)
            axis_f, envelop_freq2 = ecosound.core.tools.resample_1D_array(
                minigram.axis_frequencies,
                envelop_freq,
                resolution=self.resolution_freq,
                kind=self.interp)
            if sum(envelop_freq2)==0:
                print('here')
            if sum(envelop_time2)==0:
                print('here')
            # Frequency envelop features
            features_envelop_freq = self.envelop_features(axis_f, envelop_freq2)
            if debug:
                axis_orig = minigram.axis_frequencies
                envelop_orig = envelop_freq
                axis_interp = axis_f
                envelop_interp = envelop_freq2
                features = features_envelop_freq
                title = 'Frequency envelop'
                SpectrogramFeatures._plot_envelop_features(axis_orig,
                                           envelop_orig,
                                           axis_interp,
                                           envelop_interp,
                                           features,
                                           title=title)
            # Time envelop features
            features_envelop_time = self.envelop_features(axis_t, envelop_time2)
            if debug:
                axis_orig = minigram.axis_times
                envelop_orig = envelop_time
                axis_interp = axis_t
                envelop_interp = envelop_time2
                features = features_envelop_time
                title = 'Time envelop'
                SpectrogramFeatures._plot_envelop_features(axis_orig,
                                           envelop_orig,
                                           axis_interp,
                                           envelop_interp,
                                           features,
                                           title=title)
            # Amplitude modulation features
            # TO DO
            # Full spectrogram matrix features
            adjusted_bounds = [features_envelop_time['pct5_position'].values[0],
                               features_envelop_time['pct95_position'].values[0],
                               features_envelop_freq['pct5_position'].values[0],
                               features_envelop_freq['pct95_position'].values[0],
                               ]
            features_spectrogram, frequency_points = self.spectrogram_features(minigram,adjusted_bounds=adjusted_bounds)
            if debug:
                SpectrogramFeatures._plot_spectrogram_features(minigram,
                                               features_spectrogram,
                                               adjusted_bounds,
                                               frequency_points,
                                               title='spectrogram features')
            # stack all features
            tmp = pd.DataFrame({
                'uuid': [annot['uuid']],
                # from frequency envelop
                'freq_peak': features_envelop_freq['peak_position'],
                'freq_bandwidth': features_envelop_freq['length'],
                'freq_bandwidth90': features_envelop_freq['length_90'],
                'freq_pct5': features_envelop_freq['pct5_position'],
                'freq_pct25': features_envelop_freq['pct25_position'],
                'freq_pct50': features_envelop_freq['pct50_position'],
                'freq_pct75': features_envelop_freq['pct75_position'],
                'freq_pct95': features_envelop_freq['pct95_position'],
                'freq_iqr': features_envelop_freq['IQR'],
                'freq_asymmetry': features_envelop_freq['asymmetry'],
                'freq_concentration': features_envelop_freq['concentration'],
                'freq_std': features_envelop_freq['std'],
                'freq_kurtosis': features_envelop_freq['kurtosis'],
                'freq_skewness': features_envelop_freq['skewness'],
                'freq_entropy': features_envelop_freq['entropy'],
                'freq_flatness': features_envelop_freq['flatness'],
                'freq_roughness': features_envelop_freq['roughness'],
                'freq_centroid': features_envelop_freq['centroid'],
                # from full spectrogram
                'freq_overall_peak': features_spectrogram['freq_peak'],
                'freq_median_mean': features_spectrogram['freq_median_mean'],
                'freq_median_std': features_spectrogram['freq_median_std'],
                'freq_entropy_mean': features_spectrogram['freq_entropy_mean'],
                'freq_entropy_std': features_spectrogram['freq_entropy_std'],
                'freq_upsweep_mean': features_spectrogram['freq_upsweep_mean'],
                'freq_upsweep_fraction': features_spectrogram['freq_upsweep_fraction'],
                'snr': features_spectrogram['snr'],
                # from time envelop
                'time_peak_sec': features_envelop_time['peak_position'],
                'time_peak_perc': features_envelop_time['peak_position_relative'],
                'time_duration': features_envelop_time['length'],
                'time_duration90': features_envelop_time['length_90'],
                'time_pct5': features_envelop_time['pct5_position'],
                'time_pct25': features_envelop_time['pct25_position'],
                'time_pct50': features_envelop_time['pct50_position'],
                'time_pct75': features_envelop_time['pct75_position'],
                'time_pct95': features_envelop_time['pct95_position'],
                'time_iqr': features_envelop_time['IQR'],
                'time_asymmetry': features_envelop_time['asymmetry'],
                'time_concentration': features_envelop_time['concentration'],
                'time_std': features_envelop_time['std'],
                'time_kurtosis': features_envelop_time['kurtosis'],
                'time_skewness': features_envelop_time['skewness'],
                'time_entropy': features_envelop_time['entropy'],
                'time_flatness': features_envelop_time['flatness'],
                'time_roughness': features_envelop_time['roughness'],
                'time_centroid': features_envelop_time['centroid'],
                })
            # stack features for each annotation
            if index == 0:
                features = tmp
                features_name = list(features.columns)
                features_name.remove('uuid')
            else:
                features = pd.concat([features, tmp], ignore_index=False)
        # merge with annotation fields
        annotations.data.set_index('uuid', inplace=True, drop=False)
        features.set_index('uuid', inplace=True, drop=True)
        meas = pd.concat([annotations.data, features], axis=1, join='inner')
        meas.reset_index(drop=True, inplace=True)
        # create Measurement object
        measurements = Measurement(measurer_name=self.name,
                                   measurer_version=self.version,
                                   measurements_name=features_name)
        measurements.data = meas
        return measurements

    def compute(self, spectro, annotations, debug=False, verbose=False, use_dask=False):
        """ Compute spectrogram features.

        Goes through each annotation and compute features from the spectrogram.

        Parameters
        ----------
        spectro : ecosound Spectrogram object
            Spectrogram of the recording to analyze.
        annotations : ecosound Annotation object
            Annotations of the sounds to measure. Can be from manual analysis
            or from an automatic detector.
        use_dask : bool, optional
            If True, run the measurer in parallele using Dask. The default is
            False.
        debug : bool, optional
            Displays figures for each annotation with the spectrogram, spectral
            and time envelopes, and tables with all associated measurements.
            The default is False.
        verbose : bool, optional
            Prints in the console the annotation being processed. The default
            is False.

        Returns
        -------
        measurements : ecosound Measurement object
            Measurement object containing the measurements appended to the
            original annotation fields. Measurements are in the .data data
            frame. Metadata with mearurer name, version and measurements names
            are in the .metadata datafreame.

        """
        self._prerun_check(spectro, annotations)

        #init
        features = self._init_dataframe()
        features_name = list(features.columns)
        # loop through each annotation
        df_list=[]
        for index, annot in annotations.data.iterrows():
            if verbose:
                print('processing annotation ', index, annot['time_min_offset'], '-' ,annot['time_max_offset'])
            #if index == 555:
            #    print('debug')
            # feature for 1 annot
            # tmp = self.compute_single_annot(annot, spectro, debug, verbose)
            # # stack features for each annotation
            # features = pd.concat([features, tmp], ignore_index=False)
            # feature for 1 annot
            if use_dask:
                df = delayed(self.compute_single_annot)(annot, spectro, debug)
            else:
                df = self.compute_single_annot(annot, spectro, debug)
            # stack features for each annotation
            df_list.append(df)
        if use_dask:
            features = delayed(pd.concat)(df_list, ignore_index=False)
            #features.visualize('measuremnets')
            features = features.compute()
        else:
            features = pd.concat(df_list, ignore_index=False)
        # merge with annotation fields
        annotations.data.set_index('uuid', inplace=True, drop=False)
        features.set_index('uuid', inplace=True, drop=True)
        meas = pd.concat([annotations.data, features], axis=1, join='inner')
        meas.reset_index(drop=True, inplace=True)
        # create Measurement object
        measurements = Measurement(measurer_name=self.name,
                                   measurer_version=self.version,
                                   measurements_name=features_name)
        measurements.data = meas
        return measurements

    def _init_dataframe(self):
        tmp = pd.DataFrame({
                'uuid': [],
                # from frequency envelop
                'freq_peak': [],
                'freq_bandwidth': [],
                'freq_bandwidth90': [],
                'freq_pct5': [],
                'freq_pct25': [],
                'freq_pct50': [],
                'freq_pct75': [],
                'freq_pct95': [],
                'freq_iqr': [],
                'freq_asymmetry': [],
                'freq_concentration': [],
                'freq_std': [],
                'freq_kurtosis': [],
                'freq_skewness': [],
                'freq_entropy': [],
                'freq_flatness': [],
                'freq_roughness': [],
                'freq_centroid': [],
                # from full spectrogram
                'freq_overall_peak': [],
                'freq_median_mean': [],
                'freq_median_std': [],
                'freq_entropy_mean': [],
                'freq_entropy_std': [],
                'freq_upsweep_mean': [],
                'freq_upsweep_fraction': [],
                'snr': [],
                # from time envelop
                'time_peak_sec': [],
                'time_peak_perc': [],
                'time_duration': [],
                'time_duration90': [],
                'time_pct5': [],
                'time_pct25': [],
                'time_pct50': [],
                'time_pct75': [],
                'time_pct95': [],
                'time_iqr': [],
                'time_asymmetry': [],
                'time_concentration': [],
                'time_std': [],
                'time_kurtosis': [],
                'time_skewness': [],
                'time_entropy': [],
                'time_flatness': [],
                'time_roughness': [],
                'time_centroid': [],
                })
        return tmp
    def compute_single_annot(self, annot, spectro, debug):
            tmin = annot['time_min_offset']
            tmax = annot['time_max_offset']
            fmin = annot['frequency_min']
            fmax = annot['frequency_max']
            # extract minmgram for that detection
            minigram = spectro.crop(frequency_min=fmin,
                                    frequency_max=fmax,
                                    time_min=tmin,
                                    time_max=tmax)
            if minigram.spectrogram.any():
                # extract time and frequency envelops
                envelop_time, envelop_freq = SpectrogramFeatures.get_envelops(minigram,
                                                                normalize=True)
                # interpolate each envelop
                axis_t, envelop_time2 = ecosound.core.tools.resample_1D_array(
                    minigram.axis_times,
                    envelop_time,
                    resolution=self.resolution_time,
                    kind=self.interp)
                axis_f, envelop_freq2 = ecosound.core.tools.resample_1D_array(
                    minigram.axis_frequencies,
                    envelop_freq,
                    resolution=self.resolution_freq,
                    kind=self.interp)
                # if sum(envelop_freq2)==0:
                #     print('here')
                # if sum(envelop_time2)==0:
                #     print('here')
                # Frequency envelop features
                features_envelop_freq = self.envelop_features(axis_f, envelop_freq2)
                if debug:
                    axis_orig = minigram.axis_frequencies
                    envelop_orig = envelop_freq
                    axis_interp = axis_f
                    envelop_interp = envelop_freq2
                    features = features_envelop_freq
                    title = 'Frequency envelop'
                    SpectrogramFeatures._plot_envelop_features(axis_orig,
                                               envelop_orig,
                                               axis_interp,
                                               envelop_interp,
                                               features,
                                               title=title)
                # Time envelop features
                features_envelop_time = self.envelop_features(axis_t, envelop_time2)
                if debug:
                    axis_orig = minigram.axis_times
                    envelop_orig = envelop_time
                    axis_interp = axis_t
                    envelop_interp = envelop_time2
                    features = features_envelop_time
                    title = 'Time envelop'
                    SpectrogramFeatures._plot_envelop_features(axis_orig,
                                               envelop_orig,
                                               axis_interp,
                                               envelop_interp,
                                               features,
                                               title=title)
                # Amplitude modulation features
                # TO DO
                # Full spectrogram matrix features
                adjusted_bounds = [features_envelop_time['pct5_position'].values[0],
                                   features_envelop_time['pct95_position'].values[0],
                                   features_envelop_freq['pct5_position'].values[0],
                                   features_envelop_freq['pct95_position'].values[0],
                                   ]
                features_spectrogram, frequency_points = self.spectrogram_features(minigram,adjusted_bounds=adjusted_bounds)
                if debug:
                    SpectrogramFeatures._plot_spectrogram_features(minigram,
                                                   features_spectrogram,
                                                   adjusted_bounds,
                                                   frequency_points,
                                                   title='spectrogram features')
                # stack all features
                tmp = pd.DataFrame({
                    'uuid': [annot['uuid']],
                    # from frequency envelop
                    'freq_peak': features_envelop_freq['peak_position'],
                    'freq_bandwidth': features_envelop_freq['length'],
                    'freq_bandwidth90': features_envelop_freq['length_90'],
                    'freq_pct5': features_envelop_freq['pct5_position'],
                    'freq_pct25': features_envelop_freq['pct25_position'],
                    'freq_pct50': features_envelop_freq['pct50_position'],
                    'freq_pct75': features_envelop_freq['pct75_position'],
                    'freq_pct95': features_envelop_freq['pct95_position'],
                    'freq_iqr': features_envelop_freq['IQR'],
                    'freq_asymmetry': features_envelop_freq['asymmetry'],
                    'freq_concentration': features_envelop_freq['concentration'],
                    'freq_std': features_envelop_freq['std'],
                    'freq_kurtosis': features_envelop_freq['kurtosis'],
                    'freq_skewness': features_envelop_freq['skewness'],
                    'freq_entropy': features_envelop_freq['entropy'],
                    'freq_flatness': features_envelop_freq['flatness'],
                    'freq_roughness': features_envelop_freq['roughness'],
                    'freq_centroid': features_envelop_freq['centroid'],
                    # from full spectrogram
                    'freq_overall_peak': features_spectrogram['freq_peak'],
                    'freq_median_mean': features_spectrogram['freq_median_mean'],
                    'freq_median_std': features_spectrogram['freq_median_std'],
                    'freq_entropy_mean': features_spectrogram['freq_entropy_mean'],
                    'freq_entropy_std': features_spectrogram['freq_entropy_std'],
                    'freq_upsweep_mean': features_spectrogram['freq_upsweep_mean'],
                    'freq_upsweep_fraction': features_spectrogram['freq_upsweep_fraction'],
                    'snr': features_spectrogram['snr'],
                    # from time envelop
                    'time_peak_sec': features_envelop_time['peak_position'],
                    'time_peak_perc': features_envelop_time['peak_position_relative'],
                    'time_duration': features_envelop_time['length'],
                    'time_duration90': features_envelop_time['length_90'],
                    'time_pct5': features_envelop_time['pct5_position'],
                    'time_pct25': features_envelop_time['pct25_position'],
                    'time_pct50': features_envelop_time['pct50_position'],
                    'time_pct75': features_envelop_time['pct75_position'],
                    'time_pct95': features_envelop_time['pct95_position'],
                    'time_iqr': features_envelop_time['IQR'],
                    'time_asymmetry': features_envelop_time['asymmetry'],
                    'time_concentration': features_envelop_time['concentration'],
                    'time_std': features_envelop_time['std'],
                    'time_kurtosis': features_envelop_time['kurtosis'],
                    'time_skewness': features_envelop_time['skewness'],
                    'time_entropy': features_envelop_time['entropy'],
                    'time_flatness': features_envelop_time['flatness'],
                    'time_roughness': features_envelop_time['roughness'],
                    'time_centroid': features_envelop_time['centroid'],
                    })
            else:
                tmp = pd.DataFrame({
                    'uuid': [annot['uuid']],
                    # from frequency envelop
                    'freq_peak': np.nan,
                    'freq_bandwidth': np.nan,
                    'freq_bandwidth90': np.nan,
                    'freq_pct5': np.nan,
                    'freq_pct25': np.nan,
                    'freq_pct50': np.nan,
                    'freq_pct75': np.nan,
                    'freq_pct95': np.nan,
                    'freq_iqr': np.nan,
                    'freq_asymmetry': np.nan,
                    'freq_concentration': np.nan,
                    'freq_std': np.nan,
                    'freq_kurtosis': np.nan,
                    'freq_skewness': np.nan,
                    'freq_entropy': np.nan,
                    'freq_flatness': np.nan,
                    'freq_roughness': np.nan,
                    'freq_centroid': np.nan,
                    # from full spectrogram
                    'freq_overall_peak': np.nan,
                    'freq_median_mean': np.nan,
                    'freq_median_std': np.nan,
                    'freq_entropy_mean': np.nan,
                    'freq_entropy_std': np.nan,
                    'freq_upsweep_mean': np.nan,
                    'freq_upsweep_fraction': np.nan,
                    'snr': np.nan,
                    # from time envelop
                    'time_peak_sec': np.nan,
                    'time_peak_perc': np.nan,
                    'time_duration': np.nan,
                    'time_duration90': np.nan,
                    'time_pct5': np.nan,
                    'time_pct25': np.nan,
                    'time_pct50': np.nan,
                    'time_pct75': np.nan,
                    'time_pct95': np.nan,
                    'time_iqr': np.nan,
                    'time_asymmetry': np.nan,
                    'time_concentration': np.nan,
                    'time_std': np.nan,
                    'time_kurtosis': np.nan,
                    'time_skewness': np.nan,
                    'time_entropy': np.nan,
                    'time_flatness': np.nan,
                    'time_roughness': np.nan,
                    'time_centroid': np.nan,
                    })
            return tmp

    def envelop_features(self, axis, values):
        """Extract fetaures from time or frequency envelop.

        These measurements are mostly based on Mellinger and Bradbury, 2007:
        Mellinger, D.K. and J.W. Bradbury. 2007. Acoustic measurement of marine
        mammal sounds in noisy environments. Proceedings of the Second
        International Conference on Underwater Acoustic Measurements:
        Technologies and Results, Heraklion, Greece, pp. 273-280. ftp://ftp.
        pmel.noaa.gov/newport/mellinger/papers/Mellinger+Bradbury07-Bioacoustic
        MeasurementInNoise-UAM,Crete.pdf.

        Measurements include:
            1- peak_position
            2- peak_position_relative
            3- length
            4- length_90
            5- pct5_position
            6- pct25_position
            7- pct50_position
            8- pct75_position
            9- pct95_position
           10- IQR
           11- asymmetry
           12- concentration
           13- std
           14- kurtosis
           15- skewness
           16- entropy
           17- flatness
           18- roughness
           19- centroid

        Parameters
        ----------
        axis : numpy array
            axis of the envelope in Hz or seconds.
        values : numpy array
            time of frequency envelope. Has the same length as axis.

        Returns
        -------
        features : pandas dataframe
            Dataframe with measurmenets of the envelope.

        """
        # peak
        peak_value, peak_position_unit, peak_position_relative = SpectrogramFeatures.peak(values, axis)
        # Position of percentiles
        percentiles_value = [5, 25, 50, 75, 95]
        percentiles_position = SpectrogramFeatures.percentiles_position(values, percentiles_value, axis=axis)
        # Inter quartile range
        inter_quart_range = percentiles_position['75'] - percentiles_position['25']
        # Asymetry
        asymmetry = SpectrogramFeatures.asymmetry(percentiles_position['25'], percentiles_position['50'], percentiles_position['75'])
        # duration/width
        length = SpectrogramFeatures.length(values, axis[1]-axis[0])
        # duration/width containing 90% of magnitude
        length_90 = percentiles_position['95'] - percentiles_position['5']
        # concentration
        concentration_unit = SpectrogramFeatures.concentration(values, axis)
        # standard deviation
        std = np.std(values)  # feat
        # kusrtosis
        kurt = kurtosis(values)  # feat
        # skewness
        skewness = skew(values)  # feat
        # entropy
        entropy = ecosound.core.tools.entropy(values)
        # flatness - spectral flatness (0: tone, 1: white noise) (see seewave)
        flatness = SpectrogramFeatures.flatness(values)
        # roughness or total curvature of a curve
        roughness = SpectrogramFeatures.roughness(values)
        # centroid
        centroid = SpectrogramFeatures.centroid(values, axis)
        # gather all feature into DataFrame
        features = pd.DataFrame({
            'peak_position': [peak_position_unit],
            'peak_position_relative': [peak_position_relative],
            'length': [length],
            'length_90': [length_90],
            'pct5_position': [percentiles_position['5']],
            'pct25_position': [percentiles_position['25']],
            'pct50_position': [percentiles_position['50']],
            'pct75_position': [percentiles_position['75']],
            'pct95_position': [percentiles_position['95']],
            'IQR': [inter_quart_range],
            'asymmetry': [asymmetry],
            'concentration': [concentration_unit],
            'std': [std],
            'kurtosis': [kurt],
            'skewness': [skewness],
            'entropy': [entropy],
            'flatness': [flatness],
            'roughness': [roughness],
            'centroid': [centroid],
            })
        return features

    def spectrogram_features(self, minigram1, adjusted_bounds=None):
        """Extract fetaures from the spectrogram.

        These measurements are mostly based on Mellinger and Bradbury, 2007:
        Mellinger, D.K. and J.W. Bradbury. 2007. Acoustic measurement of marine
        mammal sounds in noisy environments. Proceedings of the Second
        International Conference on Underwater Acoustic Measurements:
        Technologies and Results, Heraklion, Greece, pp. 273-280. ftp://ftp.
        pmel.noaa.gov/newport/mellinger/papers/Mellinger+Bradbury07-Bioacoustic
        MeasurementInNoise-UAM,Crete.pdf.

        Measurements include:
           1- freq_peak
           2- freq_median_mean
           3- freq_median_std
           4- freq_entropy_mean
           5- freq_entropy_std
           6- freq_upsweep_mean
           7- freq_upsweep_fraction
           8- snr

        Parameters
        ----------
        minigram1 : ecosound Spectrogram object
            Spectrogram of the sound to analyse.
        adjusted_bounds : list, optional
            List with defining the 90% energy time-frequency window for the
            measurmenets.
            adjusted_bounds = [Time min., Time max., Freq. min., Freq. max.].
            Times is seconds, frequencies in Hz. The default is None.

        Returns
        -------
        features : pandas dataframe
            dataframe with spectrogram measuremnets.
        frequency_points : pandas dataframe
            Dataframe with the median and peak frequency vectors with their
            time axis vector. Only used for plotting and debugging purposes.

        """
        if adjusted_bounds:
            minigram = minigram1.crop(time_min=adjusted_bounds[0],
                                      time_max=adjusted_bounds[1],
                                      frequency_min=adjusted_bounds[2],
                                      frequency_max=adjusted_bounds[3],
                                      inplace=False,
                                      )
        else:
            minigram = minigram1
        spectro = minigram.spectrogram.transpose()
        # Spectrum for each time framee
        peak_f = []
        peak_amp = []
        median_f = []
        entropy_agg = []
        if spectro.shape[1] > 1: #  must be at least 1 bin of bandwidth
            #root4_magnitude = []
            for spectrum in spectro:
                axis_f, spectrum2 = ecosound.core.tools.resample_1D_array(
                    minigram.axis_frequencies,
                    spectrum,
                    resolution=self.resolution_freq,
                    kind=self.interp)
                if sum(spectrum)>0:
                    # track peak frequency
                    peak_value, peak_position, _ = SpectrogramFeatures.peak(spectrum2, axis_f)
                    peak_amp.append(peak_value)
                    peak_f.append(peak_position)
                    # track median frequency
                    pct50_position = SpectrogramFeatures.percentiles_position(spectrum2,[50],axis_f)['50']
                    median_f.append(pct50_position)
                    # entropy
                    entropy_agg.append(ecosound.core.tools.entropy(spectrum))
                    #root4_magnitude.append(np.power(np.sum(spectrum2), 1/4))
            if len(median_f) > 1:
                # overall frequency peak
                _, freq_peak, _ = SpectrogramFeatures.peak(peak_amp, peak_f)
                # mean of median frequency track
                freq_median_mean = np.nanmean(median_f)
                # standard deviation of median frequency track
                freq_median_std = np.nanstd(median_f)
                # mean of spectral entropy track
                freq_entropy_mean = np.nanmean(entropy_agg)
                # mean of spectral entropy track
                freq_entropy_std = np.nanstd(entropy_agg)
                # Upsweep mean/fraction
                upsweep_mean, upsweep_fraction = SpectrogramFeatures.upsweep_index(median_f)
            else:
                freq_peak = np.nan
                freq_median_mean = np.nan
                freq_median_std = np.nan
                freq_entropy_mean = np.nan
                freq_entropy_std = np.nan
                upsweep_mean = np.nan
                upsweep_fraction = np.nan
        elif spectro.shape[1] == 1: # only 1 bin of bandwidth
            freq_peak = minigram.axis_frequencies[0]
            freq_median_mean = minigram.axis_frequencies[0]
            freq_median_std = 0
            freq_entropy_mean = np.nan
            freq_entropy_std = np.nan
            upsweep_mean = 0
            upsweep_fraction = 0
        else:
            freq_peak = np.nan
            freq_median_mean = np.nan
            freq_median_std = np.nan
            freq_entropy_mean = np.nan
            freq_entropy_std = np.nan
            upsweep_mean = np.nan
            upsweep_fraction = np.nan
        # signal to noise ratio
        snr = SpectrogramFeatures.snr(spectro)
        # FM features
        # med_freq_offset = np.dot((median_f -  np.mean(median_f)),root4_magnitude)
        # gather all feature into DataFrame
        features = pd.DataFrame({
            'freq_peak': [freq_peak],
            'freq_median_mean': [freq_median_mean],
            'freq_median_std': [freq_median_std],
            'freq_entropy_mean': [freq_entropy_mean],
            'freq_entropy_std': [freq_entropy_std],
            'freq_upsweep_mean': [upsweep_mean],
            'freq_upsweep_fraction': [upsweep_fraction],
            'snr': [snr],
            })
        frequency_points = pd.DataFrame({
            'axis_times': [minigram.axis_times+adjusted_bounds[0]],
            'freq_median': [median_f],
            'freq_peak': [peak_f],
            })
        return features, frequency_points

    @staticmethod
    def get_envelops(minigram, normalize=False):
        """Extract time and frequency envelop from spectrogram."""
        envelop_freq = np.sum(minigram.spectrogram, axis=1)
        envelop_time = np.sum(minigram.spectrogram, axis=0)
        if normalize:
            envelop_freq = envelop_freq/sum(envelop_freq)
            envelop_time = envelop_time/sum(envelop_time)
        return envelop_time, envelop_freq

    @staticmethod
    def _plot_envelop_features(axis_orig, envelop_orig, axis_interp, envelop_interp, features, title):
        """Plot envelope along with measurments table."""
        # plot - for debuging
        fig, ax = plt.subplots(1, 2, constrained_layout=True)
        ax[0].plot(axis_interp, envelop_interp, '.r')
        ax[0].plot(axis_orig, envelop_orig, '.g')
        ax[0].legend(['Interpolated', 'Original'])
        ax[0].grid()
        table = ax[1].table(cellText=features.values.T,
                            rowLabels=features.columns,
                            loc='center',
                            colWidths=[0.8,0.4])
        table.set_fontsize(20)
        ax[1].axis('off')
        fig.suptitle(title)
        fig.patch.set_visible(False)

    @staticmethod
    def _plot_spectrogram_features(minigram, features, adjusted_bounds, frequency_points, title=''):
        """"Plot spectrogram along with measurments table."""
        # plot - for debuging
        fig, ax = plt.subplots(1, 2, constrained_layout=True)
        ax[0].pcolormesh(minigram.axis_times,
                         minigram.axis_frequencies,
                         minigram.spectrogram,
                         cmap='jet',
                         vmin=np.percentile(minigram.spectrogram, 50),
                         vmax=np.percentile(minigram.spectrogram, 99.9)
                         )
        # ax[0].grid()
        ax[0].add_patch(Rectangle((adjusted_bounds[0], adjusted_bounds[2]),
                                  adjusted_bounds[1]-adjusted_bounds[0],
                                  adjusted_bounds[3]-adjusted_bounds[2],
                                  linewidth=2,
                                  edgecolor='white',
                                  facecolor='white',
                                  fill=False,
                                  alpha=0.8,
                                  )
                        )
        ax[0].plot(frequency_points['axis_times'].values[0],
                   frequency_points['freq_median'].values[0],
                   'xr')
        ax[0].plot(frequency_points['axis_times'].values[0],
                   frequency_points['freq_peak'].values[0],
                   'xk')
        ax[0].legend(['Median frequency', 'Peak frequency'])
        table = ax[1].table(cellText=features.values.T,
                            rowLabels=features.columns,
                            loc='center',
                            colWidths=[0.8, 0.4]
                            )
        table.set_fontsize(20)
        ax[1].axis('off')
        fig.suptitle(title)
        fig.patch.set_visible(False)

    @staticmethod
    @njit
    def length(array, resolution):
        """Duration/bandwidth of a time/frequency envelop."""
        return len(array)*resolution

    @staticmethod
    def peak(array, axis):
        """Return peak value, poistion and relative position of a
        time/frequency envelop."""
        peak_value = np.amax(array)
        idxmax = np.where(array == peak_value)[0][0]
        peak_position_unit = axis[idxmax]
        peak_position_relative = (idxmax/len(array))*100
        return peak_value, peak_position_unit, peak_position_relative

    @staticmethod
    def percentiles_position_old(array, percentiles, axis=None):
        """Provide position of a percentile in an array of values.

        Parameters
        ----------
        array : numpy array
            array with values.
        percentiles : list
            List with the percentiles to "find" (e.g. [50, 75]).
        axis : numpy array, optional
            array with axis for the array values. The default is None.

        Returns
        -------
        pct_position : dict
            Dictionary with position of the percentile. Dict keys are the
            values of the percentiles requested (e.g. pct_position['50']).
        """
        if axis is None:
            axis = range(0, len(array), 1)
        pct_position = dict()
        values_sum = np.sum(array)
        values_cumsum = np.cumsum(array)
        for pct in percentiles:
            pct_val = pct/100*values_sum
            pct_val_idx = np.where(values_cumsum > pct_val)[0][0]
            pct_val_unit = axis[pct_val_idx]
            pct_position[str(pct)] = pct_val_unit
        return pct_position

    @staticmethod
    def percentiles_position(array, percentiles, axis=None):
        """Provide position of a percentile in an array of values.

        Parameters
        ----------
        array : numpy array
            array with values.
        percentiles : list
            List with the percentiles to "find" (e.g. [50, 75]).
        axis : numpy array, optional
            array with axis for the array values. The default is None.

        Returns
        -------
        pct_position : dict
            Dictionary with position of the percentile. Dict keys are the
            values of the percentiles requested (e.g. pct_position['50']).
        """
        if axis is None:
            axis = range(0, len(array), 1)
        pct_position = dict()
        values_sum = np.sum(array)
        values_cumsum = np.cumsum(array)
        for pct in percentiles:
            pct_val = pct/100*values_sum
            pct_val_idx = np.where(values_cumsum > pct_val)[0][0]
            pct_val_unit = axis[pct_val_idx]
            pct_position[str(pct)] = pct_val_unit
        return pct_position

    @staticmethod
    @njit
    def asymmetry(pct25, pct50, pct75):
        """Calculate envelope assymetry."""
        return (pct25+pct75-(2*pct50))/(pct25+pct75)  # feat

    @staticmethod
    def concentration(array, axis):
        """Calculate envelope concentration."""
        sort_idx = np.argsort(-array)
        values_sorted = array[sort_idx]
        axis_sorted = axis[sort_idx]
        idx_pct50 = SpectrogramFeatures.percentiles_position(values_sorted, [50])['50']
        idx_pct50 = np.max([idx_pct50, 1])  # in case idx50 == 0
        unit_min = np.min(axis_sorted[0:idx_pct50])
        unit_max = np.max(axis_sorted[0:idx_pct50])
        concentration_unit = unit_max - unit_min  # feat
        return concentration_unit

    @staticmethod
    @njit
    def flatness(array):
        """Calculate envelope flatness."""
        # normalize and add 1 to account for zero values
        array = array/max(array)+1
        # arithmetic mean
        arithmetic_mean = np.mean(array)
        # geometric mean
        n = len(array)
        multiply = np.prod(array)
        geometric_mean = (multiply)**(1/n)
        #geometric_mean = gmean(array)
        return geometric_mean/arithmetic_mean

    @staticmethod
    def roughness(array):
        """Calculate envelope roughness."""
        array_norm = array/max(array)
        deriv2 = ecosound.core.tools.derivative_1d(array_norm, order=2)
        return np.sum(np.power(deriv2, 2))

    @staticmethod
    def centroid(array, axis):
        """Calculate envelope centroid."""
        return np.dot(axis, array) / np.sum(array)  # feat

    @staticmethod
    def upsweep_index(array):
        """Calculate envelope upsweep mean and upsweep fraction."""
        freq_median_delta = np.subtract(array[1:], array[0:-1])
        upsweep_mean = np.mean(freq_median_delta)
        upsweep_fraction = len(np.where(freq_median_delta >= 0)[0]) / len(freq_median_delta)
        return upsweep_mean, upsweep_fraction

    @staticmethod
    def snr(array):
        """Calculate signal to noise ratio."""
        sig = np.amax(array)
        noise = np.percentile(array, 25)
        if noise > 0:
            snr = 10*np.log10(sig/noise)
        else:
            #snr = 10*np.log10(sig)  #feat
            snr = np.nan  #feat
        return snr



# -*- coding: utf-8 -*-
"""
Created on Sun Mar  7 21:45:01 2021

@author: xavier.mouy
"""
import pandas as pd
import numpy as np; np.random.seed(0)
import seaborn as sns; sns.set_theme()
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

sns.set(rc={'axes.facecolor':'white', 'figure.facecolor':'white'})


#import altair as alt
infile = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\Kanes.etal_2021\GLMMdata_centeredSPL_AddedDatetimeColumn.csv'

# load and clean up data
data = pd.read_csv(infile)
data['date_time'] = pd.to_datetime(data['date_time'], format="'%d-%b-%Y %H:%M:%S'") # Convert date
data.drop(columns=['date','season','diel_per','SPL_100Hz_meanrem'], inplace=True)
data.set_index("date_time", inplace=True)

# add time and hour
data['Hour'] = pd.to_datetime(data.index).hour
#data['Day'] = data.index.map(lambda x: x.floor("d").tz_localize(tz=None))
data['Day'] = data.index.map(lambda x: x.strftime("%Y-%m-%d"))
#data['Day'] = data.index.map(lambda x: x.strftime("%d %b %Y"))


data_2D = data.reset_index(drop=True)
data_2D = data_2D.pivot_table(index='Hour', columns='Day',values='WSD_pres', aggfunc=np.max)
data_2D = data_2D.replace([np.inf, -np.inf,0,np.nan], np.nan)
fig, ax = plt.subplots(1)
ax = sns.heatmap(data_2D,
                 vmin=0,
                 vmax=1,
                 cmap='Greys',
                 cbar=False,
                 linewidth=0,
                 linecolor='k',
                 square=False,
                 )
ax.invert_yaxis()
plt.yticks(np.arange(0,25,2),np.arange(0,25,2), rotation=0, fontsize="12", va="center")
plt.xticks(fontsize="12")
ax.patch.set_edgecolor('black')
ax.patch.set_linewidth('2')
ax.set_ylabel('Hour of day')
ax.set_xlabel('')
fig.autofmt_xdate()
ax.grid(True, color='k', linestyle=':', linewidth=1)




# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:41:54 2020

@author: xavier.mouy

"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
#from ecosound.core.measurement import Measurement
from ecosound.detection.detector_builder import DetectorFactory
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import time

## Input paraneters ##########################################################

single_channel_file = r"../ecosound/resources/67674121.181018013806.wav"
#single_channel_file = r"../ecosound/resources/JASCOAMARHYDROPHONE742_20140913T084017.774Z.wav"
single_channel_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\DFO_snake-island_rca-in_20181017\audio_data\67674121.181018040806.wav"

# Spectrogram parameters
frame = 0.0625 #3000
nfft = 0.0853 # 4096
step = 0.01 # 500
#ovlp = 2500
fmin = 0
fmax = 1000
window_type = 'hann'

# start and stop time of wavfile to analyze
t1 = 197 #22#24
t2 = 217 #40#24#40
## ###########################################################################
tic = time.perf_counter()

# load audio data
sound = Sound(single_channel_file)
sound.read(channel=0, chunk=[t1, t2], unit='sec', detrend=True)

# Calculates  spectrogram
spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='sec')
spectro.compute(sound, dB=True, use_dask=True, dask_chunks=40)

# Crop unused frequencies
spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)

# Denoise
spectro.denoise('median_equalizer', window_duration=3, use_dask=True, dask_chunks=(2048,1000), inplace=True)

# Detector
detector = DetectorFactory('BlobDetector', use_dask=True, dask_chunks=(2048,2000), kernel_duration=0.1, kernel_bandwidth=300, threshold=10, duration_min=0.05, bandwidth_min=40)
detections = detector.run(spectro, debug=False)

toc = time.perf_counter()
print(f"Executed in {toc - tic:0.4f} seconds")

# Plot
graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000)
graph.add_data(sound)
graph.add_annotation(detections, panel=0, color='grey',label='Detections')
graph.add_data(spectro)
graph.add_annotation(detections, panel=1,color='black',label='Detections')
graph.colormap = 'binary'
#graph.colormap = 'jet'
graph.show()




## To test the .crop method
#detecSpectro = spectro.crop(time_min=2,time_max=10, inplace=False)
#detecSpectro = spectro.crop(time_max=10, inplace=False)
#detecSpectro = spectro.crop(frequency_min=50, inplace=False)
#detecSpectro = spectro.crop(frequency_max=800,inplace=False)
# detecSpectro = spectro.crop(frequency_min=0,frequency_max=600,time_min=10,time_max=10.3, inplace=False)
# graph = GrapherFactory('SoundPlotter', title='Detection', frequency_max=1000)
# graph.add_data(detecSpectro)
# graph.show()






# -*- coding: utf-8 -*-
"""
Created on Wed May 13 13:44:29 2020

@author: xavier.mouy
"""
from time import sleep, perf_counter
from dask import delayed, compute, visualize

# from dask.distributed import Client, LocalCluster
# cluster = LocalCluster()
# client = Client(cluster,processes=False)


def inc(x):
    sleep(1)
    return x + 1

def add(x, y):
    sleep(1)
    return x + y

tic = perf_counter()

data = range(1,80,1)
# Sequential code
results = []
for x in data:
    y = delayed(inc)(x)
    results.append(y)
    
total = delayed(sum)(results)
C=total.compute()
#total.visualize('dsds.png')

toc = perf_counter()
print(f"Dexecuted in {toc - tic:0.4f} seconds")


# -*- coding: utf-8 -*-
"""
Created on Fri May  8 15:54:29 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation
from ecosound.detection.detector_builder import DetectorFactory
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import ecosound.core.tools
import time
import os
import pickle
import platform
import numpy as np
import configparser
import argparse

def load_config_file(configfile):    
    # Loads config  files
    cfg = configparser.ConfigParser()
    cfg.read(configfile)    
    channel = int(cfg['AUDIO']['Channel'])
    LowPassFilter_Hz = float(cfg['AUDIO']['LowPassFilter_Hz'])
    KurtosisFrame_Sec = float(cfg['DETECTOR']['KurtosisFrame_Sec'])
    KurtosisThreshold = float(cfg['DETECTOR']['KurtosisThreshold'])
    KurtosisDelta_Sec = float(cfg['DETECTOR']['KurtosisDelta_Sec'])
    MaxDtwDist = float(cfg['CLASSIFICATION']['MaxDtwDist'])
    if type(LowPassFilter_Hz) != list:
        LowPassFilter_Hz = [LowPassFilter_Hz]
    config = ({'channel': channel,
               'LowPassFilter_Hz': LowPassFilter_Hz,
               'KurtosisFrame_Sec': KurtosisFrame_Sec,
               'KurtosisThreshold': KurtosisThreshold,
               'KurtosisDelta_Sec': KurtosisDelta_Sec,
               'MaxDtwDist': MaxDtwDist
               })
    return config
    
def run_detector(infile, outdir, classif_model=None, deployment_file=None):
    ## Input paraneters ##########################################################   
    
    
    # Spectrogram parameters
    frame = 0.0625 #3000
    nfft = 0.0853 # 4096
    step = 0.01 # 5
    fmin = 0
    fmax = 1000
    window_type = 'hann'

    # start and stop time of wavfile to analyze
    #t1 = 0 # 24
    #t2 = 60 # 40
    ## ###########################################################################
    outfile = os.path.join(outdir, os.path.split(infile)[1]+'.nc')
    
    if os.path.exists(outfile) is False:
        # load audio data
        sound = Sound(infile)
        #sound.read(channel=0, chunk=[t1, t2], unit='sec')
        sound.read(channel=0, unit='sec')
        # Calculates  spectrogram
        print('Spectrogram')
        spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='sec')
        spectro.compute(sound, dB=True, use_dask=True, dask_chunks=100)
        # Crop unused frequencies
        spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)
        # Denoise
        print('Denoise')
        spectro.denoise('median_equalizer',
                        window_duration=3,
                        use_dask=True,
                        dask_chunks= (50,50000),#'auto',#(87,10000),
                        inplace=True)
        # Detector
        print('Detector')
        file_timestamp = ecosound.core.tools.filename_to_datetime(infile)[0]
        detector = DetectorFactory('BlobDetector',
                                   kernel_duration=0.1,
                                   kernel_bandwidth=300,
                                   threshold=10,
                                   duration_min=0.05,
                                   bandwidth_min=40)
        detections = detector.run(spectro,
                                  start_time=file_timestamp,
                                  use_dask=True,
                                  dask_chunks=(4096,50000),#'auto',
                                  debug=False)
        # Maasurements
        print('Measurements')
        spectro_features = MeasurerFactory('SpectrogramFeatures', resolution_time=0.001, resolution_freq=0.1, interp='linear')
        measurements = spectro_features.compute(spectro,
                                                detections,
                                                debug=False,
                                                verbose=False,
                                                use_dask=True)
        
        # Add metadata
        if deployment_file:
            measurements.insert_metadata(deployment_file)
        
        # Add file informations
        file_name = os.path.splitext(os.path.basename(infile))[0]
        file_dir = os.path.dirname(infile)
        file_ext = os.path.splitext(infile)[1]
        measurements.insert_values(operator_name=platform.uname().node,
                                   audio_file_name=file_name,
                                   audio_file_dir=file_dir,
                                   audio_file_extension=file_ext,
                                   audio_file_start_date= ecosound.core.tools.filename_to_datetime(infile)[0]
                                   )
    
    
    
        # Classification
        print('Classification')
        if classif_model:
            features = classif_model['features']
            model = classif_model['model']
            Norm_mean = classif_model['normalization_mean']
            Norm_std = classif_model['normalization_std']
            classes_encoder = classif_model['classes']
            # data dataframe
            data = measurements.data
            n1=len(data)
            # drop observations/rows with NaNs
            data = data.replace([np.inf, -np.inf], np.nan)
            data.dropna(subset=features, axis=0, how='any', thresh=None, inplace=True)
            n2=len(data)
            print('Deleted observations (due to NaNs): ' + str(n1-n2))
            # Classification - predictions
            X = data[features]
            X = (X-Norm_mean)/Norm_std
            pred_class = model.predict(X)
            pred_prob = model.predict_proba(X)
            pred_prob = pred_prob[range(0,len(pred_class)),pred_class]
            # Relabel
            for index, row in classes_encoder.iterrows():
                pred_class = [row['label'] if i==row['ID'] else i for i in pred_class]
            # update measurements
            data['label_class'] = pred_class
            data['confidence'] = pred_prob

        
        # sort detections by ascending start date/time
        data.sort_values('time_min_offset',axis=0,ascending=True,inplace=True)
        # save result as NetCDF file
        print('Saving')
        measurements.data = data
        measurements.to_netcdf(outfile)
    else:
        print('Recording already processed.')


def main():
    indir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\DFO_snake-island_rca-in_20181017\audio_data'
    outdir=r'C:\Users\xavier.mouy\Desktop\test'    
    deployment_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\April_July2019_1342218252\deployment_info.csv'
    classif_model_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Classification\bkp\RF300_model.sav'
    ext='.wav' 
    
    # load classif model
    classif_model = pickle.load(open(classif_model_file, 'rb'))

    # list files to process
    files = ecosound.core.tools.list_files(indir,
                                            ext,
                                            recursive=False,
                                            case_sensitive=True)
    # process each file
    for idx,  file in enumerate(files):
        print(str(idx)+r'/'+str(len(files))+': '+ file)
        tic = time.perf_counter()
        run_detector(file, outdir, classif_model=classif_model, deployment_file=deployment_file)
        toc = time.perf_counter()
        print(f"Executed in {toc - tic:0.4f} seconds")

if __name__ == "__main__":
    main()
